"""
This module provides simple save/restore functionality for PVs, similar to
the autosave module in synApps for IOCs but (obviously) via Channel Access.

Request & Save file formats are designed to be compatible with synApps autosave.

Use of this module requires the pyparsing parser framework. 
The Debian/Ubuntu package is "python-pyparsing"
The web site is http://pyparsing.wikispaces.com/

"""
from . import save_restore

AutoSaver = save_restore.AutoSaver
restore_pvs = save_restore.restore_pvs
save_pvs = save_restore.save_pvs



#
# example of using a connection callback that will be called
# for any change in connection status

import epics
import time
import pvnames



import sys
write = sys.stdout.write
def onConnectionChange(pvname=None,  **kws):
    write('ca connection status changed:  %s %s\n' % ( pvname, repr(kws)))
    
chid = epics.ca.create_channel(pvnames.long_pv, callback=onConnectionChange)

write('Now, restart simulation IOC -- waiting, watching values and connection changes:\n')
t0 = time.time()
while time.time()-t0 < 15:
    time.sleep(0.001)



"""This script tests using EPICS CA and Python threads together

Based on code from  Friedrich Schotte, NIH, modified by Matt Newville
20-Apr-2010
"""

import time
import epics
import sys
import pvnames 

write = sys.stdout.write
flush = sys.stdout.flush

write('initial put: \n')
epics.caput(pvnames.motor1, -2.0)
epics.caput(pvnames.motor2, 33.0)

write('sleep...')
time.sleep(2.0)
flush()

write('now put with wait: \n')
flush()

epics.caput(pvnames.motor2, -20.0, wait=True)
write('done with move 1\n')
flush()

epics.caput(pvnames.motor2, 20.0, wait=True)
write('done with move 2\n')

epics.caput(pvnames.motor2, -20.0, wait=True)
write('done with move 3\n')

epics.caput(pvnames.motor2, 20.0, wait=True)
write('done with move 4\n')



#!/usr/bin/env python

import epics
import sys
try:
    from PySide.QtGui import QWidget, QLabel, QLineEdit, QGridLayout, QApplication
except:
    from PyQt4.QtGui import QWidget, QLabel, QLineEdit, QGridLayout, QApplication

from epics.utils import bytes2str

class PVText(QLabel):
    def __init__(self, pvname,  **kws):
        QLabel.__init__(self, '', **kws)
        self.pv = None
        self.cb_index = None

    def SetPV(self, pvname):
        if self.pv is not None and self.cb_index is not None:
            self.pv.remove_callback(self.cb_index)

        self.pv = epics.PV(bytes2str(pvname))
        self.setText(self.pv.get(as_string=True))
        self.cb_index = self.pv.add_callback(self.onPVChange)

    def onPVChange(self, pvname=None, char_value=None, **kws):
        self.setText(char_value)

class PVLineEdit(QLineEdit):
    def __init__(self, pvname=None,  **kws):
        QLineEdit.__init__(self, **kws)
        self.returnPressed.connect(self.onReturn)
        self.pv = None
        self.cb_index = None
        if pvname is not None:
            self.SetPV(pvname)

    def SetPV(self, pvname):
        if self.pv is not None and self.cb_index is not None:
            self.pv.remove_callback(self.cb_index)

        self.pv = epics.PV(bytes2str(pvname))
        self.cb_index = self.pv.add_callback(self.onPVChange)

    def onPVChange(self, pvname=None, char_value=None, **kws):
        self.setText(char_value)

    def onReturn(self):
        self.pv.put(bytes2str(self.text()))

class PVProbe(QWidget):
    def __init__(self, parent=None):
        QWidget.__init__(self, parent)
        self.setWindowTitle("PyQt4 PV Probe:")

        self.pv1name = QLineEdit()
        self.pv2name = QLineEdit()
        self.value   = PVText(None)
        self.pvedit  = PVLineEdit()

        grid = QGridLayout()
        grid.addWidget(QLabel("PV1 Name (Read-only):"),   0, 0)
        grid.addWidget(QLabel("PV1 Value (Read-only):"),  1, 0)
        grid.addWidget(QLabel("PV2 Name: (Read-write):"), 2, 0)
        grid.addWidget(QLabel("PV2 Value (Read-write):"), 3, 0)
        grid.addWidget(self.pv1name,  0, 1)
        grid.addWidget(self.value,    1, 1)
        grid.addWidget(self.pv2name,  2, 1)
        grid.addWidget(self.pvedit,   3, 1)

        self.pv1name.returnPressed.connect(self.onPV1NameReturn)
        self.pv2name.returnPressed.connect(self.onPV2NameReturn)

        self.setLayout(grid)

    def onPV1NameReturn(self):
        self.value.SetPV(self.pv1name.text())

    def onPV2NameReturn(self):
        self.pvedit.SetPV(self.pv2name.text())

if __name__ == '__main__':
    app = QApplication(sys.argv)
    probe = PVProbe()
    probe.show()
    sys.exit(app.exec_())



import time
import gc
import os

import epics


import pvnames

pvlist = pvnames.updating_pvlist

def show_memory():
    gc.collect()
    if os.name == 'nt':
        return 'Windows memory usage?? pid=%i' % os.getpid()
    f = open("/proc/%i/statm" % os.getpid())
    mem = f.readline().split()
    f.close()
    return 'Memory: VmSize = %i kB  /  VmRss = %i kB' %( int(mem[0])*4 , int(mem[1])*4)

N_new = 0
def get_callback(pvname=None,value=None,**kw):
    global N_new
    N_new = N_new + 1

def monitor_events(t = 60.0):
    print( 'Processing PV requests:')
    t0 = time.time()
    endtime = t0 + t
    global N_new
    nnotify = 10
    while time.time() < endtime:
        epics.ca.pend_event(0.01)
        if N_new >= nnotify:
            print( "Saw %i changes in %.3f seconds:  %s" % (N_new,  time.time()-t0, show_memory()))
            N_new = 0
            t0 = time.time()

def run(t=10.0):
    pvs = [epics.PV(i, callback=get_callback) for i in pvlist]
    epics.ca.pend_io(1.0)
    for p in pvs: p.get()
    print( 'Monitoring %i PVs'  % len(pvs))
    monitor_events(t=t)

    print( 'Destroying PVs: ')
    for i in pvs:
        i.disconnect()
    print( epics.ca._cache.keys())
    epics.ca.show_cache()
    epics.ca.poll(0.01, 10.0)
    time.sleep(1.0)

for i in range(4):
    print( "==run #  ", i+1)
    run(t=15)

print( 'memory leak test complete.')



import epics
import time
def onChanges(pvname=None, value=None, char_value=None, **kw):
    print('PV Changed! ', pvname, char_value, time.ctime())

mypv = epics.get_pv('PyTest:ao1.VAL', callback=onChanges)

print('Now wait for changes')
expire_time = time.time() + 60.
while time.time() < expire_time:
    time.sleep(0.01)
print('Done.')



from __future__ import print_function
import epics
import time
import multiprocessing as mp
import threading

import pvnames
PVN1 = pvnames.double_pv # 'Py:ao2'
PVN2 = pvnames.double_pv2 # 'Py:ao3'

def subprocess(*args):
    print('==subprocess==', args)
    mypvs = [epics.get_pv(pvname) for pvname in args]

    for i in range(10):
        time.sleep(0.750)
        out = [(p.pvname, p.get(as_string=True)) for p in mypvs]
        out = ', '.join(["%s=%s" % o for o in out])
        print('==sub (%d): %s' % (i, out))

def test_mpprocess():
    def monitor(pvname=None, char_value=None, **kwargs):
        print('--main:monitor %s=%s' % (pvname, char_value))

    print('--main:')
    pv1 = epics.get_pv(PVN1)
    print('--main:init %s=%s' % (PVN1, pv1.get()))
    pv1.add_callback(callback=monitor)

    try:
        proc1 = epics.CAProcess(target=subprocess,
                                args=(PVN1, PVN2))
        proc1.start()
        proc1.join()
    except KeyboardInterrupt:
        print('--main: killing subprocess')
        proc1.terminate()

    print('--main: subprocess complete')
    time.sleep(0.5)
    print('--main:final %s=%s' % (PVN1, pv1.get()))



import time
import sys

import epics

import gc

import pvnames

import os
def show_memory():
    gc.collect()
    if os.name == 'nt':
        return 'Windows memory usage?? pid=%i' % os.getpid()

    f = open("/proc/%i/statm" % os.getpid())
    mem = f.readline().split()
    f.close()
    sys.stdout.write('Memory: VmSize = %i kB  /  VmRss = %i kB\n' %( int(mem[0])*4 , int(mem[1])*4))

def get_callback(pvname=None, char_value=None,**kw):
    sys.stdout.write('OnGet %s: %s\n'  % (pvname, char_value))

def monitor_events(t = 10.0):
    sys.stdout.write('Processing PV requests:\n')
    t0 = time.time()
    while time.time()-t0 < t :
        epics.ca.poll()
x
def round():
    sys.stdout.write('== Creating some PVs\n ')
    pvs = []
    for field in ('VAL','DESC', 'OFF','FOFF', 'HLM','LLM','SET'):
        pvs.append(epics.PV("%s.%s" % (pvnames.motor1,field), callback=get_callback) )
    epics.ca.poll()

    for p in pvs: p.connect()
    monitor_events(t=4.)

    epics.ca.show_cache()
    sys.stdout.write('Destroying PVs:\n ')
    sys.stdout.flush()
    for i in pvs:  i.disconnect()

    monitor_events(t=0.5)

for i in range(20):
    round()
    show_memory()

epics.ca.pend_io(1.0)



#!/usr/bin/env python
"""
Epics scan record
"""
from .. import Device, poll
import threading

NUM_POSITIONERS = 4
NUM_TRIGGERS    = 4
NUM_DETECTORS   = 70

class Scan(Device):
    """
    A Device representing an Epics sscan record.
    """

    attrs = ('VAL', 'SMSG', 'CMND', 'NPTS', 'EXSC', 'NAME', 'PDLY',
             'PAUS', 'CPT', 'DDLY')

    pos_attrs = ('PV', 'SP', 'EP', 'SI', 'CP', 'WD', 'PA', 'AR', 'SM')
    trig_attrs = ('PV', 'NV')

    _alias = {'device':      'P1PV',
              'start':       'P1SP',
              'end':         'P1EP',
              'step':        'P1SI',
              'table':       'P1PA',
              'absrel':      'P1AR',
              'mode':        'P1SM',
              'npts':        'NPTS',
              'execute':     'EXSC',
              'trigger':     'T1PV',
              'pause':       'PAUS',
              'current_point':  'CPT'}

    def __init__(self, name, **kwargs):
        """
        Initialize the scan.

        name: The name of the scan record.
        """
        attrs = list(self.attrs)
        for i in range(1, NUM_POSITIONERS+1):
            for a in self.pos_attrs:
                attrs.append('P%i%s' % (i, a))
        for i in range(1, NUM_TRIGGERS+1):
            for a in self.trig_attrs:
                attrs.append('T%i%s' % (i, a))
        for i in range(1, NUM_DETECTORS+1):
            attrs.append('D%2.2iPV' % i)

        self.waitSemaphore = threading.Semaphore(0)
        Device.__init__(self, name, delim='.', attrs=attrs, **kwargs)
        for attr, pv in Scan._alias.items():
            self.add_pv('%s.%s' % (name,pv), attr)

        # make sure this is really a sscan!
        rectype = self.get('RTYP')
        if rectype != 'sscan':
            raise ScanException("%s is not an Epics Scan" % name)

        self.put('SMSG', '')

    def run(self, wait=False, timeout=86400):
        """
        Execute the scan, optionally waiting for completion

        Arguments
        ---------
        wait     whether to wait for completion, True/False (default False)
        timeout  maximum time to wait in seconds, default=86400 (1 day).

        """
        self.put('EXSC', 1, wait=wait, timeout=timeout)

    def _onDone(self, **kwargs):
        if kwargs['value'] == 0:
            self.waitSemaphore.release()

    def reset(self):
        """Reset scan, clearing positioners, detectors, triggers"""
        self.put('NPTS', 0)
        for i in range(1, NUM_TRIGGERS+1):
            self.clear_trigger(i)
        for i in range(1, NUM_POSITIONERS+1):
            self.clear_positioner(i)
        for i in range(1, NUM_DETECTORS+1):
            self.clear_detector(i)
        poll(1.e-3, 1.0)

    def _print(self):
        print('PV = %s' % self.get('P1PV'))
        print('SP = %s' % self.get('P1SP'))
        print('EP = %s' % self.get('P1EP'))
        print('NPTS = %s' % self.get('NPTS'))
        print('T  = %s' % self.get('T1PV'))


    def clear_detector(self, idet=1):
        """completely clear a detector

        Arguments
        ---------
          idet    index of detector (1 through 70, default 1)
        """
        self.put("D%2.2iPV" % idet, '')
        poll(1.e-3, 1.0)

    def add_detector(self, detector):
        """add a detector to a scan definition

        Arguments
        ---------
          detector   name of detector pv

        Returns
        -------
         idet  index of detector set
        """
        idet = None
        for _idet in range(1, NUM_DETECTORS+1):
            poll(1.e-3, 1.0)
            if len(self.get('D%2.2iPV' % _idet)) < 2:
                idet = _idet
                break
        if idet is None:
            raise ScanException("%i Detectors already defined." % (NUM_DETECTORS))
        self.put("D%2.2iPV" % idet, detector, wait=True)
        return idet

    def clear_trigger(self, itrig=1):
        """completely clear a trigger

        Arguments
        ---------
          itrig    index of trigger (1 through 4, default 1)
        """
        self.put("T%iPV" % itrig, '')
        poll(1.e-3, 1.0)

    def add_trigger(self, trigger, value=1.0):
        """add a trigger to a scan definition

        Arguments
        ---------
          trigger   name of trigger pv
          value     value to send to trigger (default 1.0)

        Returns
        -------
           itrig  index of trigger set
        """
        itrig = None
        for _itrig in range(1, NUM_TRIGGERS+1):
            poll(1.e-3, 1.0)
            if len(self.get('T%iPV' % _itrig)) < 2:
                itrig = _itrig
                break
        if itrig is None:
            raise ScanException("%i Triggers already defined." % (NUM_TRIGGERS))

        self.put("T%iPV" % itrig, trigger, wait=True)
        self.put("T%iCD" % itrig, value, wait=True)
        return itrig


    def clear_positioner(self, ipos=1):
        """completely clear a positioner

        Arguments
        ---------
          ipos    index of positioner (1 through 4, default 1)
        """
        for attr in self.pos_attrs:
            nulval = 0
            if attr == 'PV': nulval = ''
            if attr == 'PA': nulval = [0]
            self.put("P%i%s" % (ipos, attr), nulval)
        self.put("R%iPV" % ipos, '')
        poll(1.e-3, 1.0)

    def add_positioner(self, drive, readback=None,
                       start=None, stop=None, step=None,
                       center=None, width=None,
                       mode='linear', absolute=True, array=None):
        """add a positioner to a scan definition

        Arguments
        ----------
         drive     name of drive pv
         readback  name of readback pv (defaults to .RBV if drive ends in .VAL)
         mode      positioner mode ('linear', 'table', fly', default 'linear')
         absolute  whether to use absolute values (True/False, default True)
         start     start value
         stop      stop value
         step      step value
         center    center value
         width     width value
         array     array of values for table or fly mode

        Returns
        -------
         ipos  index of positioner set

        """
        ipos = None
        for _ipos in range(1, NUM_POSITIONERS+1):
            poll(1.e-3, 1.0)
            if len(self.get('P%iPV' % _ipos)) < 2:
                ipos = _ipos
                break
        if ipos is None:
            raise ScanException("%i Positioners already defined." % (NUM_POSITIONERS))

        self.put('P%iPV' % ipos, drive, wait=True)
        if readback is None and drive.endswith('.VAL'):
            readback = drive[:-4] + '.RBV'
        if readback is not None:
            self.put('R%iPV' % ipos, readback)

        # set relative/absolute
        if absolute:
            self.put('P%iAR' % ipos, 0)
        else:
            self.put('P%iAR' % ipos, 1)

        # set mode
        smode = 0
        if mode.lower().startswith('table'):
            smode = 1
        elif mode.lower().startswith('fly'):
            smode = 2
        self.put('P%iSM' % ipos, smode)

        # start, stop, step, center, width
        if start is not None:
            self.put('P%iSP' % ipos, start)
        if stop is not None:
            self.put('P%iEP' % ipos, stop)
        if step is not None:
            self.put('P%iSI' % ipos, step)
        if center is not None:
            self.put('P%iCP' % ipos, center)
        if width is not None:
            self.put('P%iWD' % ipos, width)

        # table or fly mode
        if smode in (1, 2) and array is not None:
            self.put('P%iPA' % ipos, array)
        poll(1.e-3, 1.0)
        return ipos

    def set_positioner(self, ipos, drive=None, readback=None,
                       start=None, stop=None, step=None,
                       center=None, width=None,
                       mode=None, absolute=None, array=None):
        """change a positioner setting in a scan definition
        all settings are optional, and will leave other settings unchanged

        Arguments
        ----------
         drive     name of drive pv
         readback  name of readback pv
         mode      positioner mode ('linear', 'table', fly', default 'linear')
         absolute  whether to use absolute values (True/False, default True)
         start     start value
         stop      stop value
         step      step value
         center    center value
         width     width value
         array     array of values for table or fly mode

        Notes
        -----
         This allows changing a scan, for example:

             s = Scan('XXX:scan1')
             ipos1 = s.add_positioner('XXX:m1.VAL', start=-1, stop=1, step=0.1)
             ....

             s.run()

         Then changing the scan definition with

            s.set_positioner(ipos1, start=0, stop=0.2, step=0.01)
            s.run()
        """
        if ipos is None:
            raise ScanException("must give positioner index")

        if drive is not None:
            self.put('P%iPV' % ipos, drive)
        if readback is not None:
            self.put('R%iPV' % ipos, readback)
        if start is not None:
            self.put('P%iSP' % ipos, start)
        if stop is not None:
            self.put('P%iEP' % ipos, stop)
        if step is not None:
            self.put('P%iSI' % ipos, step)
        if center is not None:
            self.put('P%iCP' % ipos, center)
        if width is not None:
            self.put('P%iWD' % ipos, width)
        if array is not None:
            self.put('P%iPA' % ipos, array)

        if absolute is not None:
            if absolute:
                self.put('P%iAR' % ipos, 0)
            else:
                self.put('P%iAR' % ipos, 1)

        if mode is not None:
            smode = 0
            if mode.lower().startswith('table'):
                smode = 1
            elif mode.lower().startswith('fly'):
                smode = 2
            self.put('P%iSM' % ipos, smode)
        poll(1.e-3, 1.0)


    def after_scan(self, mode):
        """set after scan mode"""
        self.put("PASM", mode, wait=True)

    def positioner_delay(self, pdelay):
        """set positioner delay in seconds"""
        self.put("PDLY", pdelay, wait=True)

    def detector_delay(self, pdelay):
        """set detector delay in seconds"""
        self.put("DDLY", pdelay, wait=True)

class ScanException(Exception):
    """ raised to indicate a problem with a scan"""
    def __init__(self, msg, *args):
        Exception.__init__(self, *args)
        self.msg = msg
    def __str__(self):
        return str(self.msg)



#!/usr/bin/env python 
"""Epics Support for
Stanford Research Systems 570 current amplifier
"""
from .. import Device

VALID_STEPS = [1, 2, 5, 10, 20, 50, 100, 200, 500]
VALID_UNITS = ['pA/V', 'nA/V','uA/V', 'mA/V']

class SRS570(Device):
    """ 
    SRS (Stanford Research Systems) 570 current amplifier
    """

    attrs = ('sens_num', 'sens_unit', 'offset_num', 'offset_unit',
             'offset_sign', 'offset_on', 'off_u_put', 'bias_put',
             'gain_mode', 'filter_type', 'invert_on', 'init.PROC')
    
    _nonpvs = ('_prefix', '_pvs', '_delim', '_nchan', '_chans')
    
    def __init__(self, prefix):
        Device.__init__(self, prefix, delim='',
                        attrs=self.attrs, mutable=False)
        self.initialize()

    def initialize(self, bias=0, gain_mode=0, filter_type=0,
                   invert=False):
        """set initial values"""
        inv_val = 0
        if invert: inv_val = 1
        self.put('gain_mode', gain_mode) # 0 = low noise
        self.put('filter_type', filter_type) # 0  no filter
        self.put('invert_on', inv_val)
        self.put('bias_put', bias)
           
    def set_sensitivity(self, value, units, offset=None, 
                        scale_offset=True):
        "set sensitivity"
        if value not in VALID_STEPS or units not in VALID_UNITS:
            print('invalid input')
            return
        
        ival = VALID_STEPS.index(value)
        uval = VALID_UNITS.index(units)

        self.put('sens_num', ival)
        self.put('sens_unit', uval)
        if scale_offset:
            # scale offset to by 0.1 x sensitivity
            # i.e, a sensitivity of 200 nA/V should 
            # set set the input offset to 20 nA.
            ioff = ival - 3
            uoff = uval
            if ioff < 0:
                ioff = ival + 6
                uoff = uval - 1
            self.put('offset_num',  ioff)
            self.put('offset_unit', uoff)
        if offset is not None:
            self.set_offset(offset)
        self.put('init.PROC', 1)

    def set_offset(self, value):
        self.put('off_u_put', value)

    def increase_sensitivity(self):
        "increase sensitivity by 1 step"
        snum  = self.get('sens_num')
        sunit = self.get('sens_unit')
        if snum == 0:
            snum = 9
            sunit = sunit - 1
            if sunit < 0:
                # was at highest sensitivity
                snum, sunit = 1, 0
        snum = snum - 1
        self.set_sensitivity(VALID_STEPS[snum], VALID_UNITS[sunit])

    def decrease_sensitivity(self):
        "decrease sensitivity by 1 step"
        snum  = self.get('sens_num')
        sunit = self.get('sens_unit')
        if snum == 8:
            snum = -1
            sunit = sunit + 1
            if sunit > 3:
                # was at lowest sensitivity
                snum, sunit = 7, 3
        snum = snum + 1
        self.set_sensitivity(VALID_STEPS[snum], VALID_UNITS[sunit])




from __future__ import print_function

import time
import epics
import pvnames
print('== Test get/put for synchronous groups')

pvs = pvnames.motor_list

chids = [epics.ca.create_channel(pvname) for pvname in pvs]

for chid in chids:
    epics.ca.connect_channel(chid)
    epics.ca.put(chid, 0)

print('Now create synch group ')
sg = epics.ca.sg_create()

data = [epics.ca.sg_get(sg, chid) for chid in chids]

print('Now change these PVs for the next 10 seconds')
time.sleep(10.0)

print('Synchronous block:')
epics.ca.sg_block(sg)
print('Done.  Values')
for pvname, dat, chid in zip(pvs, data, chids):
    print("%s = %s" % (pvname, str( epics.ca._unpack(dat, chid=chid))))

epics.ca.sg_reset(sg)

print('OK, now we will put everything back to 0 synchronously')

for chid in chids:
    epics.ca.sg_put(sg, chid, 0)
print('sg_put done, but not blocked / commited. Sleep for 5 seconds ')
time.sleep(5.0)
print('Now Go: ')
epics.ca.sg_block(sg)
print('done.')


          



#!/usr/bin/env python
# test expression parsing

import os
import sys
import time
import numpy
import ctypes
from contextlib import contextmanager
from epics import ca, dbr, caput
from epics.utils import IOENCODING

import pvnames

def _ca_connect(chid,timeout=5.0):
    n  = 0
    t0 = time.time()
    conn = 2==ca.state(chid)
    while (not conn) and (time.time()-t0 < timeout):
        ca.poll(1.e-6,1.e-4)
        conn = 2==ca.state(chid)
        n += 1
    return conn, time.time()-t0, n

def write(msg):
    sys.stdout.write('%s\n'% msg)
    sys.stdout.flush()

CONN_DAT ={}
CHANGE_DAT = {}

def onConnect(pvname=None, conn=None, chid=None,  **kws):
    write('  /// Connection status changed:  %s  %s' % (pvname, repr(kws)))
    global CONN_DAT
    CONN_DAT[pvname] = conn

def onChanges(pvname=None, value=None, **kws):
    write( '/// New Value: %s  value=%s, kw=%s' %( pvname, str(value), repr(kws)))
    global CHANGE_DAT
    CHANGE_DAT[pvname] = value

@contextmanager
def no_simulator_updates():
    '''Context manager which pauses and resumes simulator PV updating'''
    try:
        caput(pvnames.pause_pv, 1)
        yield
    finally:
        caput(pvnames.pause_pv, 0)


def test_CreateChid():
    write('Simple Test: create chid')
    chid = ca.create_channel(pvnames.double_pv)
    assert chid is not None

def test_GetNonExistentPV():
    write('Simple Test: get on a non-existent PV')
    chid = ca.create_channel('Definitely-Not-A-Real-PV')
    val = ca.get(chid)
    assert val is None

def test_CreateChid_CheckTypeCount():
    write('Simple Test: create chid, check count, type, host, and access')
    chid = ca.create_channel(pvnames.double_pv)
    ret = ca.connect_channel(chid)
    ca.pend_event(1.e-3)

    ftype  = ca.field_type(chid)
    count  = ca.element_count(chid)
    host    = ca.host_name(chid)
    rwacc = ca.access(chid)

    assert chid is not None
    assert host is not None
    assert count == 1
    assert ftype == 6
    assert rwacc == 'read/write'


def test_CreateChidWithConn():
    write('Simple Test: create chid with conn callback')
    chid = ca.create_channel(pvnames.int_pv,
                             callback=onConnect)
    val = ca.get(chid)
    global CONN_DAT
    conn = CONN_DAT.get(pvnames.int_pv, None)
    assert conn

def test_dbrName():
    write( 'DBR Type Check')
    assert dbr.Name(dbr.STRING) == 'STRING'
    assert dbr.Name(dbr.FLOAT) == 'FLOAT'
    assert dbr.Name(dbr.ENUM) == 'ENUM'
    assert dbr.Name(dbr.CTRL_CHAR) == 'CTRL_CHAR'
    assert dbr.Name(dbr.TIME_DOUBLE) == 'TIME_DOUBLE'
    assert dbr.Name(dbr.TIME_LONG) == 'TIME_LONG'

    assert dbr.Name('STRING', reverse=True) == dbr.STRING
    assert dbr.Name('DOUBLE', reverse=True) == dbr.DOUBLE
    assert dbr.Name('CTRL_ENUM', reverse=True) == dbr.CTRL_ENUM
    assert dbr.Name('TIME_LONG', reverse=True) == dbr.TIME_LONG

def test_Connect1():
    chid = ca.create_channel(pvnames.double_pv)
    conn,dt,n = _ca_connect(chid, timeout=2)
    write( 'CA Connection Test1: connect to existing PV')
    write( ' connected in %.4f sec' % (dt))
    assert conn

def test_Connected():
    pvn = pvnames.double_pv
    chid = ca.create_channel(pvn,connect=True)
    isconn = ca.isConnected(chid)
    write( 'CA test Connected (%s) = %s' % (pvn,isconn))
    assert isconn
    assert ca.state(chid) == ca.dbr.CS_CONN
    assert ca.access(chid) == 'read/write'

def test_DoubleVal():
    pvn = pvnames.double_pv
    chid = ca.create_channel(pvn,connect=True)
    cdict  = ca.get_ctrlvars(chid)
    write( 'CA testing CTRL Values for a Double (%s)'   % (pvn))
    assert 'units' in cdict
    assert 'precision' in cdict
    assert 'severity' in cdict

    assert len(ca.host_name(chid)) > 2
    assert ca.element_count(chid) == 1

    assert ca.field_type(chid) == ca.dbr.DOUBLE
    assert ca.get_precision(chid) == pvnames.double_pv_prec
    assert ca.get_ctrlvars(chid)['units'] == pvnames.double_pv_units
    assert ca.access(chid).startswith('read')

def test_UnConnected():
    write( 'CA Connection Test1: connect to non-existing PV (2sec timeout)')
    chid = ca.create_channel('impossible_pvname_certain_to_fail')
    conn, dt, n = _ca_connect(chid, timeout=2)
    assert not conn

def test_putwait():
    'test put with wait'
    pvn = pvnames.non_updating_pv
    chid = ca.create_channel(pvn, connect=True)
    o  = ca.put(chid, -1, wait=True)
    time.sleep(0.01)
    assert ca.get(chid) == -1

    ca.put(chid, 2, wait=True)
    assert ca.get(chid) == 2

def test_promote_type():
    pvn = pvnames.double_pv
    chid = ca.create_channel(pvn,connect=True)
    write( 'CA promote type (%s)' % (pvn))
    assert ca.promote_type(chid, use_time=True) == ca.dbr.TIME_DOUBLE
    assert ca.promote_type(chid, use_ctrl=True) == ca.dbr.CTRL_DOUBLE

def test_ProcPut():
    pvn  = pvnames.enum_pv
    chid = ca.create_channel(pvn, connect=True)
    write( 'CA test put to PROC Field (%s)' % (pvn))
    for input in (1, '1', 2, '2', 0, '0', 50, 1):
        ret = None
        try:
            ret = ca.put(chid, 1)
        except:
            pass
        assert ret is not None

def test_subscription_double():
    pvn = pvnames.updating_pv1
    chid = ca.create_channel(pvn,connect=True)
    cb, uarg, eventID = ca.create_subscription(chid, callback=onChanges)

    start_time = time.time()
    global CHANGE_DAT
    while time.time()-start_time < 5.0:
        time.sleep(0.01)
        if CHANGE_DAT.get(pvn, None) is not None:
            break
    val = CHANGE_DAT.get(pvn, None)
    ca.clear_subscription(eventID)
    assert val is not None

def test_subscription_custom():
    pvn = pvnames.updating_pv1
    chid = ca.create_channel(pvn, connect=True)

    global change_count
    change_count = 0

    def my_callback(pvname=None, value=None, **kws):
        write( ' Custom Callback  %s  value=%s' %(pvname, str(value)))
        global change_count
        change_count = change_count + 1

    cb, uarg, eventID = ca.create_subscription(chid, callback=my_callback)

    start_time = time.time()
    while time.time()-start_time < 2.0:
        time.sleep(0.01)

    ca.clear_subscription(eventID)
    time.sleep(0.2)
    assert change_count > 2

def test_subscription_str():
    pvn = pvnames.updating_str1
    write(" Subscription on string: %s " % pvn)
    chid = ca.create_channel(pvn,connect=True)
    cb, uarg, eventID = ca.create_subscription(chid, callback=onChanges)

    start_time = time.time()
    global CHANGE_DAT
    while time.time()-start_time < 3.0:
        time.sleep(0.01)
        ca.put(chid, "%.1f" % (time.time()-start_time) )
        if CHANGE_DAT.get(pvn, None) is not None:
            break
    assert CHANGE_DAT.get(pvn, None) is not None

def arrcallback(arrayname, array_type, length, element_type):
    """ Helper function to subscribe to a PV array and check it
        receives at least one subscription callback w/ specified type,
        length & uniform element type. Checks separately for normal,
        TIME & CTRL subscription variants. Returns the array or fails
    an assertion."""
    results = {}
    for form in [ 'normal', 'time', 'ctrl' ]:
        chid = ca.create_channel(arrayname,connect=True)
        cb, uarg, eventID = ca.create_subscription(chid, use_time=form=='time', use_ctrl=form=='ctrl', callback=onChanges)

        CHANGE_DAT.pop(arrayname, None)
        timeout=0
        # wait up to 6 seconds, if no callback probably due to simulator.py
        # not running...
        while timeout<120 and not arrayname in CHANGE_DAT:
            time.sleep(0.05)
            timeout = timeout+1
        val = CHANGE_DAT.get(arrayname, None)
        ca.clear_subscription(eventID)
        assert val is not None
        assert type(val) == array_type
        assert len(val) == length
        assert type(val[0]) == element_type
        assert all(type(e)==element_type for e in val)
        results[form] = val
    return results

def test_subscription_long_array():
    """ Check that numeric arrays callbacks successfully send correct data """
    arrcallback(pvnames.long_arr_pv, numpy.ndarray, 2048, numpy.int32)

def test_subscription_double_array():
    """ Check that double arrays callbacks successfully send correct data """
    arrcallback(pvnames.double_arr_pv, numpy.ndarray, 2048, numpy.float64)

def test_subscription_string_array():
    """ Check that string array callbacks successfully send correct data """
    results = arrcallback(pvnames.string_arr_pv, list, 128, str)
    assert len(results["normal"][0]) > 0
    assert len(results["time"][0]) > 0
    assert len(results["ctrl"][0]) > 0

def test_subscription_char_array():
    """ Check that uchar array callbacks successfully send correct data as arrays """
    arrcallback(pvnames.char_arr_pv, numpy.ndarray, 128, numpy.uint8)

def test_Values():
    write( 'CA test Values (compare 5 values with caget)')
    os.system('rm ./caget.tst')
    vals = {}
    with no_simulator_updates():
        for pvn in (pvnames.str_pv,  pvnames.int_pv,
                    pvnames.float_pv, pvnames.enum_pv,
                    pvnames.long_pv):
            os.system('caget  -n -f5 %s >> ./caget.tst' % pvn)
            chid = ca.create_channel(pvn)
            ca.connect_channel(chid)
            vals[pvn] = ca.get(chid)
        rlines = open('./caget.tst', 'r', encoding=IOENCODING).readlines()
        for line in rlines:
            pvn, sval = [i.strip() for i in line[:-1].split(' ', 1)]
            tval = str(vals[pvn])
            if pvn in (pvnames.float_pv,pvnames.double_pv):
                # use float precision!
                tval = "%.5f" % vals[pvn]
            assert tval == sval

def test_type_converions_1():
    write("CA type conversions scalars")
    pvlist = (pvnames.str_pv, pvnames.int_pv, pvnames.float_pv,
              pvnames.enum_pv,  pvnames.long_pv,  pvnames.double_pv2)
    chids = []
    with no_simulator_updates():
        for name in pvlist:
            chid = ca.create_channel(name)
            ca.connect_channel(chid)
            chids.append((chid, name))
            ca.poll(evt=0.025, iot=5.0)
        ca.poll(evt=0.05, iot=10.0)

        values = {}
        for chid, name in chids:
            values[name] = ca.get(chid, as_string=True)

        for promotion in ('ctrl', 'time'):
            for chid, pvname in chids:
                write('=== %s  chid=%s as %s' % (ca.name(chid), repr(chid),
                                                 promotion))
                time.sleep(0.01)
                if promotion == 'ctrl':
                    ntype = ca.promote_type(chid, use_ctrl=True)
                else:
                    ntype = ca.promote_type(chid, use_time=True)

                val  = ca.get(chid, ftype=ntype)
                cval = ca.get(chid, as_string=True)
                if ca.element_count(chid) > 1:
                    val = val[:12]
                assert cval == values[pvname]

def test_type_converions_2():
    write("CA type conversions arrays")
    pvlist = (pvnames.char_arr_pv,
              pvnames.long_arr_pv,
              pvnames.double_arr_pv)
    with no_simulator_updates():
        chids = []
        for name in pvlist:
            chid = ca.create_channel(name)
            ca.connect_channel(chid)
            chids.append((chid, name))
            ca.poll(evt=0.025, iot=5.0)
        ca.poll(evt=0.05, iot=10.0)

        values = {}
        for chid, name in chids:
            values[name] = ca.get(chid)
        for promotion in ('ctrl', 'time'):
            for chid, pvname in chids:
                write('=== %s  chid=%s as %s' % (ca.name(chid), repr(chid),
                                                 promotion))
                time.sleep(0.01)
                if promotion == 'ctrl':
                    ntype = ca.promote_type(chid, use_ctrl=True)
                else:
                    ntype = ca.promote_type(chid, use_time=True)

                val  = ca.get(chid, ftype=ntype)
                cval = ca.get(chid, as_string=True)
                for a, b in zip(val, values[pvname]):
                    assert a == b


def test_Array0():
    write('Array Test: get double array as numpy array, ctypes Array, and list')
    chid = ca.create_channel(pvnames.double_arrays[0])
    aval = ca.get(chid)
    cval = ca.get(chid, as_numpy=False)

    assert isinstance(aval, numpy.ndarray)
    assert len(aval) > 2
    assert isinstance(cval, ctypes.Array)
    assert len(cval) > 2
    lval = list(cval)
    assert isinstance(lval, list)
    assert len(lval) > 2
    assert lval == list(aval)

def test_xArray1():
    write('Array Test: get(wait=False) / get_complete()')
    chid = ca.create_channel(pvnames.double_arrays[0])
    val0 = ca.get(chid)
    aval = ca.get(chid, wait=False)
    assert aval is  None
    val1 = ca.get_complete(chid)
    assert all(val0 == val1)

def test_xArray2():
    write('Array Test: get fewer than max vals using ca.get(count=0)')
    chid = ca.create_channel(pvnames.double_arrays[0])
    maxpts = ca.element_count(chid)
    npts = int(max(2, maxpts/2.3 - 1))
    write('max points is %s' % (maxpts, ))
    dat = numpy.random.normal(size=npts)
    write('setting array to a length of npts=%s' % (npts, ))
    ca.put(chid, dat)
    out1 = ca.get(chid)
    assert isinstance(out1, numpy.ndarray)
    assert len(out1) == npts
    out2 = ca.get(chid, count=0)
    assert isinstance(out2, numpy.ndarray)
    assert len(out2) == npts

def test_xArray3():
    write('Array Test: get char array as string')
    chid = ca.create_channel(pvnames.char_arrays[0])
    val = ca.get(chid)
    assert isinstance(val, numpy.ndarray)
    char_val = ca.get(chid, as_string=True)
    assert isinstance(char_val, str)
    conv = ''.join([chr(i) for i in val])
    assert conv == char_val



#!/usr/bin/python
#
# simple PV Probe application

import wx
import sys
import epics
from epics.wx import EpicsFunction, DelayedEpicsCallback

class PVDisplay(wx.Frame):
    def __init__(self, pvname, parent=None, **kwds):
        wx.Frame.__init__(self, parent, wx.ID_ANY, '',
                         wx.DefaultPosition, wx.Size(-1,-1),**kwds)
        self.SetFont(wx.Font(11,wx.SWISS,wx.NORMAL,wx.BOLD,False))
        self.pvname = pvname

        self.SetTitle("%s" % pvname)
        
        self.sizer = wx.GridBagSizer(3, 2)
        panel = wx.Panel(self)
        name      = wx.StaticText(panel, label=pvname,        size=(120, -1))
        self.val  = wx.StaticText(panel, label='unconnected', size=(200, -1))
        self.info = wx.StaticText(panel, label='-- ' ,        size=(400,300))
        self.info.SetFont(wx.Font(9,wx.SWISS,wx.NORMAL,wx.BOLD,False))
        
        self.sizer.Add(wx.StaticText(panel, label='PV: ',    size=(60, -1)),
                       (0, 0), (1, 1), wx.EXPAND, 1)
        self.sizer.Add(wx.StaticText(panel, label='Value: ',  size=(60, -1)),
                       (1, 0), (1, 1), wx.EXPAND, 1)
        self.sizer.Add(wx.StaticText(panel, label='Info: ',   size=(60, -1)),
                       (2, 0), (1, 1), wx.EXPAND, 1)
        self.sizer.Add(name,      (0, 1), (1, 1), wx.EXPAND, 1)
        self.sizer.Add(self.val,  (1, 1), (1, 1), wx.ALIGN_RIGHT|wx.EXPAND, 1)
        self.sizer.Add(self.info, (2, 1), (2, 2), wx.ALIGN_LEFT|wx.EXPAND, 1)

        panel.SetSizer(self.sizer)

        self.needs_info = None
        self.timer = wx.Timer(self)
        self.Bind(wx.EVT_TIMER, self.onTimer, self.timer)

        self.s1 = wx.BoxSizer(wx.VERTICAL)
        self.s1.Add(panel, 1, wx.EXPAND, 2)
        self.s1.Fit(self)
        self.Refresh()
        self.connect_pv()
        
    @EpicsFunction
    def connect_pv(self):
        self.pv = epics.PV(self.pvname, connection_callback=self.onConnect,
                           callback=self.onPV_value)

    @EpicsFunction
    def onTimer(self, evt):
        if self.need_info and self.pv.connected:
            self.info.SetLabel(self.pv.info)
            self.timer.Stop()
            self.needs_info = False
        
    @DelayedEpicsCallback
    def onConnect(self, **kws):
        self.need_info = True
        self.timer.Start(25)
                
    @DelayedEpicsCallback
    def onPV_value(self, name=None, char_value=None, **kws):
        if len(char_value) > 90:
            char_value = char_value[:90]
        self.val.SetLabel('   %s' % char_value)
        self.need_info = True
        self.timer.Start(25)

        
class NameCtrl(wx.TextCtrl):
    def __init__(self, panel, value='', action=None, **kws):
        self.action = action
        wx.TextCtrl.__init__(self, panel, wx.ID_ANY, value='',
                             style=wx.TE_PROCESS_ENTER, **kws)
        self.Bind(wx.EVT_CHAR, self.onChar)

    def onChar(self, event):
        if event.GetKeyCode() == wx.WXK_RETURN and \
           self.action is not None:
            self.action(wx.TextCtrl.GetValue(self).strip())
        event.Skip()
        

class ProbeFrame(wx.Frame):
    def __init__(self, parent=None, **kwds):

        wx.Frame.__init__(self, parent, wx.ID_ANY, '',
                         wx.DefaultPosition, wx.Size(-1,-1),**kwds)
        self.SetTitle("Connect to Epics Records:")

        self.SetFont(wx.Font(11,wx.SWISS,wx.NORMAL,wx.BOLD,False))

        sizer = wx.BoxSizer(wx.HORIZONTAL)
        panel = wx.Panel(self)
        label = wx.StaticText(panel, label='PV Name:')
        self.pvname = NameCtrl(panel, value='', size=(175,-1),
                               action=self.onName)

        sizer.Add(label,       0, wx.ALIGN_LEFT, 1)
        sizer.Add(self.pvname, 1, wx.EXPAND, 1)
        panel.SetSizer(sizer)
        sizer.Fit(panel)
        s = wx.BoxSizer(wx.VERTICAL)
        s.Add(panel)
        s.Fit(self)
        self.Refresh()

    def onName(self, value, wid=None, **kws):
        PVDisplay(value, parent=self).Show()

if __name__ == '__main__':
    app = wx.App(redirect=False)
    ProbeFrame().Show()
    app.MainLoop()



#!/usr/bin/env python 
"""Epics Scaler"""
from .. import Device, poll

class Scaler(Device):
    """ 
    Simple implementation of SynApps Scaler Record.   
    """
    attrs = ('CNT', 'CONT', 'TP', 'T', 'VAL')
    attr_kws = {'calc_enable': '%s_calcEnable.VAL'}
    chan_attrs = ('NM%i', 'S%i')
    calc_attrs = {'calc%i': '%s_calc%i.VAL', 'expr%i': '%s_calc%i.CALC'}
    _nonpvs = ('_prefix', '_pvs', '_delim', '_nchan', '_chans')
    
    def __init__(self, prefix, nchan=8):
        self._nchan  = nchan
        self._chans = range(1, nchan+1)
        
        attrs = list(self.attrs)
        for i in self._chans:
            for att in self.chan_attrs:
                attrs.append(att % i)
                
        Device.__init__(self, prefix, delim='.', attrs=attrs)

        for key, val in self.attr_kws.items():
            self.add_pv(val % prefix, attr= key)
            
        for i in self._chans:
            for key, val in self.calc_attrs.items():
                self.add_pv(val % (prefix, i), attr = key % i)
        self._mutable = False
        
    def AutoCountMode(self):
        "set to autocount mode"
        self.put('CONT', 1)

    def OneShotMode(self):
        "set to one shot mode"        
        self.put('CONT', 0)

    def CountTime(self, ctime):
        "set count time"
        self.put('TP', ctime)
        
    def Count(self, ctime=None, wait=False):
        "set count, with optional counttime"
        if ctime is not None:
            self.CountTime(ctime)
        self.put('CNT', 1, wait=wait)
        poll()

    def EnableCalcs(self):
        " enable calculations"
        self.put('calc_enable', 1)

    def setCalc(self, i, calc):
        "set the calculation for scaler i"
        attr = 'expr%i'  % i
        self.put(attr, calc)

    def getNames(self):
        "get all names"
        return [self.get('NM%i' % i) for i in self._chans]

    def Read(self, use_calc=False):
        "read all values"
        attr = 'S%i'
        if use_calc:
            attr = 'calc%i'
        return [self.get(attr % i) for i in self._chans]



import time
import sys
class debugtime(object):
    def __init__(self):
        self.clear()
        self.add('debugtime init')

    def clear(self):
        self.times = []

    def add(self,msg='', verbose=False):
        # print msg
        self.times.append((msg,time.time()))
        if verbose:
            sys.stdout.write("%s\n"% msg)
            
    def show(self, writer=None, clear=True):
        if writer is None:
            writer = sys.stdout.write
        writer('%s\n' % self.get_report())
        if clear:
            self.clear()
            

    def get_report(self):
        m0, t0 = self.times[0]
        tlast= t0
        out  = ["# %s  %s " % (m0,time.ctime(t0))]
        lmsg = 0
        for m, t in self.times[1:]:
            lmsg = max(lmsg, len(m))            
        m = '#      Message'
        m = m + ' '*(lmsg-len(m))

        out.append("#--------------------" + '-'*lmsg)
        out.append('%s        Delta(s)     Total(s)' % (m))
        for m,t in self.times[1:]:
            tt = t-t0
            dt = t-tlast
            if len(m)<lmsg:
                m = m + ' '*(lmsg-len(m))
            out.append(" %s   %10.3f    %10.3f" % (m,dt, tt))
            tlast = t
        return '\n'.join(out)



#
# example of using a connection callback that will be called
# for any change in connection status

import epics
import time
import sys
from  pvnames import updating_pv1, updating_pvlist
epics.ca.PREEMPTIVE_CALLBACK = True 
write = sys.stdout.write
def onConnectionChange(pvname=None, conn= None, **kws):
    write('Connection changed: %s conn=%s (%s)\n' % (pvname,  repr(conn), time.ctime()))
    sys.stdout.flush()

def onValueChange(pvname=None, value=None, host=None, **kws):
    write('Value changed: %s = %s (%s)\n' % ( pvname, repr(value), time.ctime()))
    sys.stdout.flush()


# pv1 = epics.PV(updating_pv1, 
#                 connection_callback= onConnectionChange,
#                 callback= onValueChange)
# 
pxs = [epics.PV(thispv, 
                connection_callback= onConnectionChange,
                callback= onValueChange) for thispv in updating_pvlist]

for x in pxs:
    write("%s = %s\n" % (x.pvname, x.get(as_string=True)))

write('Now waiting, watching values and connection changes:\n')
t0 = time.time()
while time.time()-t0 < 300:
    try:
        time.sleep(0.01)
    except KeyboardInterrupt:
        break
#
# write('Some value changes should have been seens\n')
# write('Now, restart the IOC:\n')
# t0 = time.time()
# while time.time()-t0 < 60:
#     time.sleep(0.01)
# 
# write('You should have seen a connection message and new values\n')


write("done!\n")

epics.ca.show_cache()




#!/usr/bin/env python
"""Epics analog input record"""
from .. import Device

class ai(Device):
    "Simple analog input device"

    attrs = ('VAL', 'EGU', 'HOPR', 'LOPR', 'PREC', 'NAME', 'DESC',
             'DTYP', 'INP', 'LINR', 'RVAL', 'ROFF', 'EGUF', 'EGUL',
             'AOFF', 'ASLO', 'ESLO', 'EOFF', 'SMOO', 'HIHI', 'LOLO',
             'HIGH', 'LOW', 'HHSV', 'LLSV', 'HSV', 'LSV', 'HYST')

    def __init__(self, prefix, **kwargs):
        if prefix.endswith('.'):
            prefix = prefix[:-1]
        Device.__init__(self, prefix, delim='.', attrs=self.attrs, **kwargs)



import time
import epics
import pvnames
import random
import numpy
import sys

def test_subarray():
    driver = epics.PV(pvnames.subarr_driver)
    sub1   = epics.PV(pvnames.subarr1)
    sub2   = epics.PV(pvnames.subarr2)
    sub3   = epics.PV(pvnames.subarr3)
    sub4   = epics.PV(pvnames.subarr4)
    s1_0 =  int(epics.caget("%s.INDX" % pvnames.subarr1))
    s2_0 =  int(epics.caget("%s.INDX" % pvnames.subarr2))
    s3_0 =  int(epics.caget("%s.INDX" % pvnames.subarr3))
    s4_0 =  int(epics.caget("%s.INDX" % pvnames.subarr4))

    s1_n =  int(epics.caget("%s.NELM" % pvnames.subarr1))
    s2_n =  int(epics.caget("%s.NELM" % pvnames.subarr2))
    s3_n =  int(epics.caget("%s.NELM" % pvnames.subarr3))
    s4_n =  int(epics.caget("%s.NELM" % pvnames.subarr4))

    npts = len(driver.get())

    for i in range(10):
        driver.put([100*random.random() for x in range(npts)])
        time.sleep(0.1)

        full = driver.get()
        sys.stdout.write("%s\n" % full)
        assert all( [all( sub1.get() == full[s1_0:s1_n+s1_0]),
                     all( sub2.get() == full[s2_0:s2_n+s2_0]),
                     all( sub3.get() == full[s3_0:s3_n+s3_0]),
                     all( sub4.get() == full[s4_0:s4_n+s4_0])
                     ])



"""
simple devices
"""
from .ai import ai
from .ao import ao
from .bi import bi
from .bo import bo

from .scaler import Scaler
from .struck import Struck
from .srs570 import SRS570
from .mca import DXP, MCA, MultiXMAP, ROI
from .scan import Scan
from .transform import Transform

from .ad_base import AD_Camera
from .ad_fileplugin import AD_FilePlugin
from .ad_image import AD_ImagePlugin
from .ad_overlay import AD_OverlayPlugin
from .ad_perkinelmer import AD_PerkinElmer

Mca = MCA




"""
Port of Xiaogiang Wang's CaChannel class to use epics.ca

Matt Newville  20-October-2010

Original Comments:

CaChannel class having identical API as of caPython/CaChannel class, 
based on PythonCA ( > 1.20.1beta2)

Author:     Xiaoqiang Wang
Created:    Sep. 22, 2008
Changes:     
"""

from epics import ca, dbr
from epics.wx import closure
import time
import types


code = '''
class CaChannelException(Exception):
    def __init__(self, status):
        self.status = str(status)
    def __str__(self):
        return self.status

class CaChannel(object):
    """CaChannel: A Python class with identical API as of caPython/CaChannel 

    Example:
        import CaChannel
        chan = CaChannel.CaChannel('catest')
        chan.searchw()
        print chan.getw()
    """
    ca_timeout = 1.0

    def __init__(self, pvName=None):
        self.pvname = pvName
        self.__chid = None
        self.__evid = None
        self.__timeout = None
        self._field_type = None
        self._element_count = None
        self._puser = None
        self._conn_state = None
        self._host_name = None
        self._raccess = None
        self._waccess = None

        self._callbacks={}

    def __del__(self):
        try:
            self.clear_event()
            self.clear_channel()
            self.flush_io()
        except:
            pass

    def version(self):
        print("CaChannel, version v03 (pyepics port of v02-11-09)")
#
# Class helper methods
#
    def setTimeout(self, timeout):
        """Set the timeout for this channel."""
        if (timeout>=0 or timeout == None):
            self.__timeout = timeout
        else:
            raise ValueError
    def getTimeout(self):
        """Retrieve the timeout set for this channel."""
        return self.__timeout


#
# *************** Channel access medthod ***************
#

#
# Connection methods
#   search_and_connect
#   search
#   clear_channel

    def search_and_connect(self, pvName, callback, *user_args):
        """Attempt to establish a connection to a process variable.
        Parameters:
            pvName: process variable name
            callback: function called when connection completes and connection
            status changes later on.
            *user_args: user provided arguments that are passed to callback when
            it is invoked.
        """
        if pvName == None:
            pvName = self.pvname
        conn_callback = closure(callback, *user_args)
        try:
            self.__chid = ca.create_channel(pvName, callback=conn_callback)
        except ca.ChannelAccessException, msg:
            raise CaChannelException(msg)

    def search(self, pvName=None):
        """Attempt to establish a connection to a process variable.
        Parameters:
            pvName: process variable name
        """
        if pvName == None:
            pvName = self.pvname
        try:
            self.__chid = ca.create_channel(pvName)
        except ca.ChannelAccessException, msg:
            raise CaChannelException, msg

    def clear_channel(self):
        """Close a channel created by one of the search functions"""
        if(self.__chid is not None):
            try:
                status = ca.clear_channel(self.__chid)
            except ca.ChannelAccessException, msg:
                raise CaChannelException,msg 

#
# Write methods
#   array_put
#   array_put_callback
#

    def _setup_put(self, value, req_type, count = None):
        if count is None:
            count = self.element_count()
        else:
            count = max(1, min(self.element_count(), count) )

        if req_type == -1:
            req_type = self.field_type()

        # single numeric value
        if (isinstance(value, int) or 
            isinstance(value, long) or 
            isinstance(value, float) or
            isinstance(value, bool)):
            pval = (CaChannel.dbr_d[req_type](value),)
        # single string value
        #   if DBR_CHAR, split into chars
        #   otherwise convert to field type
        elif isinstance(value, str):
            if req_type == ca.DBR_CHAR:
                if len(value) < count:
                    count = len(value)
                pval = [ord(x) for x in value[:count]]
            else:
                pval = (CaChannel.dbr_d[req_type](value),)
        # assumes other sequence type
        else:
            if len(value) < count:
                count = len(value)
            pval = [CaChannel.dbr_d[req_type](x) for x in value[:count]]

        return pval

    def array_put(self, value, req_type=None, count=None):
        """Write a value or array of values to a channel
        Parameters:
            value: data to be written. For multiple values use a list or tuple
            req_type: database request type. Defaults to be the native data type.
            count: number of data values to write, Defaults to be the native count.
        """
        if req_type is None: req_type = -1
        val = self._setup_put(value, req_type, count)
        try:
            ca.put(self.__chid, val, None, None, req_type)
        except ca.ChannelAccessException,msg:
            raise CaChannelException,msg

    def array_put_callback(self, value, req_type, count, callback, *user_args):
        """Write a value or array of values to a channel and execute the user
        supplied callback after the put has completed.
        Parameters:
            value: data to be written. For multiple values use a list or tuple.
            req_type: database request type. Defaults to be the native data type.
            count: number of data values to write, Defaults to be the native count.
            callback: function called when the write is completed.
            *user_args: user provided arguments that are passed to callback when
            it is invoked.
        """
        if req_type is None: req_type = -1
        val = self._setup_put(value, req_type, count)
        self._callbacks['putCB']=(callback, user_args)
        try:
            ca.put(self.__chid, val, None, self._put_callback, req_type)
        except ca.ChannelAccessException,msg:
            raise CaChannelException,msg
#
# Read methods
#   getValue
#   array_get
#   array_get_callback
#

    # Obtain read value after ECA_NORMAL is returned on an array_get(). 
    def getValue(self):
        """Return the value(s) after array_get has completed"""
        return self.val

    # Simulate with a synchronous getw function call
    def array_get(self, req_type=None, count=None):
        """Read a value or array of values from a channel. The new value is
        retrieved by a call to getValue method.
        Parameters:
            req_type: database request type. Defaults to be the native data type.
            count: number of data values to read, Defaults to be the native count.
        """
        self.val = self.getw(req_type, count)

    def array_get_callback(self, req_type, count, callback, *user_args):
        """Read a value or array of values from a channel and execute the user
        supplied callback after the get has completed.
        Parameters:
            req_type: database request type. Defaults to be the native data type.
            count: number of data values to read, Defaults to be the native count.
            callback: function called when the get is completed.
            *user_args: user provided arguments that are passed to callback when
            it is invoked.
        """
        if req_type is None: req_type = -1
        if count is None: count = 0
        self._callbacks['getCB']=(callback, user_args)
        try:
            status=ca.get(self.__chid, self._get_callback, req_type, count)
        except ca.ChannelAccessException,msg:
            raise CaChannelException,msg

#
# Event methods
#   add_masked_array_event
#   clear_event
#

    # Creates a new event id and stores it on self.__evid.  Only one event registered
    # per CaChannel object.  If an event is already registered the event is cleared
    # before registering a new event.
    def add_masked_array_event(self, req_type, count, mask, callback, *user_args):
        """Specify a callback function to be executed whenever changes occur to a PV.
        Parameters:
            req_type: database request type. Defaults to be the native data type.
            count: number of data values to read, Defaults to be the native count.
            mask: logical or of ca.DBE_VALUE, ca.DBE_LOG, ca.DBE_ALARM. Defaults to
            be ca.DBE_VALUE|ca.DBE_ALARM.
            callback: function called when the get is completed.
            *user_args: user provided arguments that are passed to callback when
            it is invoked.
        """
        if req_type is None: req_type = -1
        if count is None: count = 0
        if mask is None: mask = ca.DBE_VALUE|ca.DBE_ALARM
        if self.__evid is not None:
            self.clear_event()
            self.pend_io()
        self._callbacks['eventCB']=(callback, user_args)
        try:
            self.__evid = ca.monitor(self.__chid, self._event_callback, count, mask)
        except ca.ChannelAccessException,msg:
            raise CaChannelException,msg

    def clear_event(self):
        """Remove previously installed callback function."""
        if self.__evid is not None:
            try:
                status=ca.clear_monitor(self.__evid)
                self.__evid = None
            except ca.ChannelAccessException,msg:
               raise CaChannelException,msg 

#
# Execute methods
#   pend_io
#   pend_event
#   poll
#   flush_io
#

    def pend_io(self,timeout=None):
        """Flush the send buffer and wait until outstanding queries complete
        or the specified timeout expires.
        Parameters:
            timeout: seconds to wait
        """
        if timeout is None:
            if self.__timeout is None:
                timeout = self.ca_timeout
            else:
                timeout = self.__timeout
        status = ca.pend_io(float(timeout))
        if status != 0:
            raise CaChannelException, ca.caError._caErrorMsg[status]

    def pend_event(self,timeout=None):
        """Flush the send buffer and wait for timeout seconds.
        Parameters:
            timeout: seconds to wait
        """
        if timeout is None:
            timeout = 0.1
        status = ca.pend_event(timeout)
        # status is always ECA_TIMEOUT
        return status

    def poll(self):
        """Flush the send buffer and execute outstanding background activity."""
        status = ca.poll() 
        # status is always ECA_TIMEOUT
        return status

    def flush_io(self):
        """Flush the send buffer and does not execute outstanding background activity."""
        status = ca.flush()
        if status != 0:
            raise CaChannelException, ca.caError._caErrorMsg[status]

#
# Channel Access Macros
#   field_type
#   element_count
#   name
#   state
#   host_name
#   read_access
#   write_access
#
    def get_info(self):
        try:
            info=(self._field_type, self._element_count, self._puser,
                  self._conn_state, self._host_name, self._raccess,
                  self._waccess) = ca.ch_info(self.__chid)
        except ca.ChannelAccessException,msg:
            raise CaChannelException,msg
        return info

    def field_type(self):
        """Native field type."""
        self.get_info()
        return self._field_type

    def element_count(self):
        """Native element count."""
        self.get_info()
        return self._element_count

    def name(self):
        """Channel name specified when the channel was connected."""
        return ca.name(self.__chid)

    def state(self):
        """Current state of the connections.
        Possible channel states:
            ca.cs_never_conn    PV not found
            ca.cs_prev_conn     PV was found but unavailable
            ca.cs_conn          PV was found and available
            ca.cs_closed        PV not closed
            ca.cs_never_search  PV not searched yet
        """
        if self.__chid is None:
            return dbr.CS_NEVER_SEARCH
        else:
            self.get_info()
            return self._conn_state

    def host_name(self):
        """Host name that hosts the process variable."""
        self.get_info()
        return self._host_name

    def read_access(self):
        """Right to read the channel."""
        self.get_info()
        return self._raccess

    def write_access(self):
        """Right to write the channel."""
        self.get_info()
        return self._waccess
#
# Wait functions
#
# These functions wait for completion of the requested action.
    def searchw(self, pvName=None):
        """Attempt to establish a connection to a process variable.
        Parameters:
            pvName: process variable name
        """
        if pvName is None:
            pvName = self.pvname
        self.__chid = ca.search(pvName, None)
        if self.__timeout is None:
            timeout = self.ca_timeout
        else:
            timeout = self.__timeout
        status = ca.pend_io(timeout)
        if status != 0:
            raise CaChannelException, ca.caError._caErrorMsg[status]

    def putw(self, value, req_type=None):
        """Write a value or array of values to a channel
        Parameters:
            value: data to be written. For multiple values use a list or tuple
            req_type: database request type. Defaults to be the native data type.
        """
        if req_type is None: req_type = -1
        val = self._setup_put(value, req_type)
        ca.put(self.__chid, val, None, None, req_type)
        if self.__timeout is None:
            timeout = self.ca_timeout
        else:
            timeout = self.__timeout
        status = ca.pend_io(timeout)
        if status != 0:
            raise CaChannelException, ca.caError._caErrorMsg[status]

    def getw(self, req_type=None, count=None):
        """Return a value or array of values from a channel.
        Parameters:
            req_type: database request type. Defaults to be the native data type.
            count: number of data values to read, Defaults to be the native count.
        """
        updated = [False]
        value = [0]
        def update_value(valstat):
            if valstat is None:
                return
            try:
                value[0] = valstat[0]
            finally:
                updated[0] = True
        if req_type is None: req_type = -1
        if count is None: count = 0
        ca.get(self.__chid, update_value, req_type, count)
        if self.__timeout is None:
            timeout = self.ca_timeout
        else:
            timeout = self.__timeout
        status = ca.pend_io(timeout)
        n = 0
        while n*0.02 < timeout and not updated[0]:
            ca.pend_event(0.02)
            n+=1
        if not updated[0]:
            raise CaChannelException, ca.caError._caErrorMsg[10] # ECA_TIMEOUT
        return value[0]

#
# Callback functions
#
# These functions hook user supplied callback functions to CA extension

    def _conn_callback(self):
        try:
            callback, userArgs = self._callbacks.get('connCB')
        except:
            return
        if self.state() == 2: OP=6
        else: OP=7
        epicsArgs = (self.__chid, OP)
        callback(epicsArgs, userArgs)

    def _put_callback(self, args):
        try:
            callback, userArgs = self._callbacks.get('putCB')
        except:
            return
        epicsArgs={}
        epicsArgs['chid']=self.__chid
        epicsArgs['type']=self.field_type()
        epicsArgs['count']=self.element_count()
        epicsArgs['status']=args[1]
        callback(epicsArgs, userArgs)

    def _get_callback(self, args):
        try:
            callback, userArgs = self._callbacks.get('getCB')
        except:
            return
        epicsArgs = self._format_cb_args(args)
        callback(epicsArgs, userArgs)

    def _event_callback(self, args): 
        try:
            callback, userArgs = self._callbacks.get('eventCB')
        except:
            return
        epicsArgs = self._format_cb_args(args)
        callback(epicsArgs, userArgs)

    def _format_cb_args(self, args):
        epicsArgs={}
        epicsArgs['chid']   = self.__chid
        # dbr_type is not returned
        # use dbf_type instead
        epicsArgs['type']   = self.field_type()
        epicsArgs['count']  = self.element_count()
        # status flag is not returned,
        # args[1] is alarm status
        # assume ECA_NORMAL
        epicsArgs['status'] = 1
        if len(args)==2:          # Error
            epicsArgs['pv_value']   = args[0] # always None
            epicsArgs['status']     = args[1]
        if len(args)>=3:          # DBR_Plain
            epicsArgs['pv_value']   = args[0]
            epicsArgs['pv_severity']= args[1]
            epicsArgs['pv_status']  = args[2]
        if len(args)>=4:          # DBR_TIME, 0.0 for others
            epicsArgs['pv_seconds'] = args[3]
        if len(args)==5:
            if len(args[4])==2:   # DBR_CTRL_ENUM
                epicsArgs['pv_nostrings']   = args[4][0]
                epicsArgs['pv_statestrings']= args[4][1]
            if len(args[4])>=7:   # DBR_GR
                epicsArgs['pv_units']       = args[4][0]
                epicsArgs['pv_updislim']    = args[4][1]
                epicsArgs['pv_lodislim']    = args[4][2]
                epicsArgs['pv_upalarmlim']  = args[4][3]
                epicsArgs['pv_upwarnlim']   = args[4][4]
                epicsArgs['pv_loalarmlim']  = args[4][5]
                epicsArgs['pv_lowarnlim']   = args[4][6]
            if len(args[4])==8:   # DBR_GR_FLOAT or DBR_GR_DOUBLE
                epicsArgs['pv_precision']   = args[4][7]
            if len(args[4])>=9:   # DBR_CTRL
                epicsArgs['pv_upctrllim']   = args[4][7]
                epicsArgs['pv_loctrllim']   = args[4][8]
            if len(args[4])==10:  # DBR_CTRL_FLOAT or DBR_CTRL_DOUBLE
                epicsArgs['pv_precision']   = args[4][9]
        return epicsArgs

'''




#
# list of local pv names to use for testing


#### 1
# this pv should be a DOUBLE.  It will NOT be set, but
# you should provide the host_name, units, and precision.  It
# is assumed to have count=1
double_pv = 'PyTest:ao1'
double_pv_units = 'microns'
double_pv_prec = 4

double_pv2 = 'PyTest:ao2'

pause_pv  = 'PyTest:pause'
#### 2
# this pv should be an ENUM. It will NOT be set.
# provide the names of the ENUM states

#### Theae are PVs of the various native types
###  They will NOT be set.
str_pv   = 'PyTest:ao1.DESC'
int_pv   = 'PyTest:long2'
long_pv  = 'PyTest:long2'
float_pv = 'PyTest:ao3'
enum_pv  = 'PyTest:mbbo1'
enum_pv_strs = ['Stop', 'Start', 'Pause', 'Resume']

proc_pv = 'PyTest:ao1.PROC'

## Here are some waveform / array data PVs
long_arr_pv   = 'PyTest:long2k'
double_arr_pv = 'PyTest:double2k'
string_arr_pv = 'PyTest:string128'
# char / byte array
char_arr_pv   = 'PyTest:char128'
char_arrays   = ['PyTest:char128', 'PyTest:char2k', 'PyTest:char64k']
long_arrays   = ['PyTest:long128', 'PyTest:long2k', 'PyTest:long64k']
double_arrays   = ['PyTest:double128', 'PyTest:double2k', 'PyTest:double64k']


####
# provide a single motor prefix (to which '.VAL' and '.RBV' etc will be added)

motor_list = ['sim:mtr%d' % i for i in range(1, 7)]
motor1 = motor_list[0]
motor2 = motor_list[1]

####
#  Here, provide a PV that changes at least once very 10 seconds
updating_pv1  = 'PyTest:ao1'
updating_str1 = 'PyTest:char256'

####
#  Here, provide a list of PVs that  change at least once very 10 seconds
updating_pvlist = ['PyTest:ao1', 'PyTest:ai1', 'PyTest:long1', 'PyTest:ao2']
#### alarm test

non_updating_pv = 'PyTest:ao4'

alarm_pv = 'PyTest:long1'
alarm_comp='ge'
alarm_trippoint = 7


#### subarray test
subarr_driver = 'PyTest:wave_test'
subarr1       = 'PyTest:subArr1'
subarr2       = 'PyTest:subArr2'
subarr3       = 'PyTest:subArr3'
subarr4       = 'PyTest:subArr4'
zero_len_subarr1 = 'PyTest:ZeroLenSubArr1'


#### clear cache tests
clear_cache_enabled = 'PyTestClearCache:enabled'
clear_cache_beacons = ['PyTestClearCache:{}'.format(i) for i in range(1, 10)]



from epics import ca, dbr

import time
import debugtime
try:
    from collections import OrderedDict
except:
    from ordereddict import OrderedDict
dt = debugtime.debugtime()
def add(x):
    print(x)
    dt.add(x)

add('test of fast connection to many PVs')
pvnames = []

results = OrderedDict()

MAX_PVS = 12500

for line  in open('fastconn_pvlist.txt','r').readlines():
    if not line.startswith('#'):
        pvnames.append(line.strip())

if MAX_PVS is not None:
    pvnames = pvnames[:MAX_PVS]


add('Read PV list:  Will connect to %i PVs' % len(pvnames))
libca = ca.initialize_libca()

for name in pvnames:
    chid = ca.create_channel(name, connect=False, auto_cb=False)
    results[name] = {'chid': chid}

time.sleep(0.001)

add("created PVs with ca_create_channel")

for name in pvnames:
    ca.connect_channel(results[name]['chid'])

time.sleep(0.001)

add("connected to PVs with connect_channel")

ca.pend_event(1.e-2)

for name in pvnames:
    chid = results[name]['chid']
    val = ca.get(chid, wait=False)
    results[name]['value'] =  val

add("did ca.get(wait=False)")
ca.poll(2.e-3, 1.0)
add("ca.poll() complete")

for name in pvnames:
    results[name]['value'] = ca.get_complete(results[name]['chid'])

add("ca.get_complete() for all PVs")

f = open('fastconn_pvdata.sav', 'w')
for name, val in results.items():
    f.write("%s %s\n" % (name.strip(), val['value']))
f.close()
add("wrote PV values to disk")

dt.show()

time.sleep(0.01)
ca.poll()



#!/usr/bin/env python
# epics/wx/utils.py
"""
This is a collection of general purpose utility functions and classes,
especially useful for wx functionality
"""
import wx
import wx.lib.masked as masked

import os
import array


BAD_FILECHARS = ';~,`!%$@$&^?*#:"/|\'\\\t\r\n (){}[]<>'
GOOD_FILECHARS = '_'*len(BAD_FILECHARS)
TRANS_FILE = str.maketrans(BAD_FILECHARS, GOOD_FILECHARS)

HAS_NUMPY = False
try:
    import numpy
    HAS_NUMPY = True
except ImportError:
    pass

# some common abbrevs for wx ALIGNMENT styles
RIGHT = wx.ALIGN_RIGHT
LEFT  = wx.ALIGN_LEFT
CEN   = wx.ALIGN_CENTER
LCEN  = wx.ALIGN_LEFT
RCEN  = wx.ALIGN_RIGHT
CCEN  = wx.ALIGN_CENTER
LTEXT = wx.ST_NO_AUTORESIZE|wx.ALIGN_CENTER


def make_steps(prec=3, tmin=0, tmax=10, base=10, steps=(1, 2, 5)):
    """make a list of 'steps' to use for a numeric ComboBox
    returns a list of floats, such as
        [0.01, 0.02, 0.05, 0.10, 0.20, 0.50, 1.00, 2.00...]
    """
    steplist = []
    power = -prec
    step = tmin
    while True:
        decade = base**power
        for step in (j*decade for j in steps):
            if step > 0.99*tmin and step <= tmax and step not in steplist:
                steplist.append(step)
        if step >= tmax:
            break
        power += 1
    return steplist

def set_sizer(panel, sizer=None, style=wx.VERTICAL, fit=False):
    """ utility for setting wx Sizer  """
    if sizer is None:
        sizer = wx.BoxSizer(style)
    panel.SetAutoLayout(1)
    panel.SetSizer(sizer)
    if fit:
        sizer.Fit(panel)

def set_float(val):
    """ utility to set a floating value,
    useful for converting from strings """
    out = None
    if not val in (None, ''):
        try:
            out = float(val)
        except ValueError:
            return None
        if HAS_NUMPY:
            if numpy.isnan(out):
                out = default
        else:
            if not(out > 0) and not(out<0) and not(out==0):
                out = default
    return out

def pack(window, sizer):
    "simple wxPython pack function"
    window.SetSizer(sizer)
    sizer.Fit(window)

def add_button(parent, label, size=(-1, -1), action=None):
    "add simple button with bound action"
    thisb = wx.Button(parent, label=label, size=size)
    if callable(action):
        thisb.Bind(wx.EVT_BUTTON, action)
    return thisb

def add_menu(parent, menu, label='', text='', action=None):
    "add submenu"
    wid = wx.NewId()
    menu.Append(wid, label, text)
    if callable(action):
        parent.Bind(wx.EVT_MENU, action, id=wid)

def popup(parent, message, title, style=None):
    """
    generic popup message dialog, returns
    output of MessageDialog.ShowModal()
    """
    if style is None:
        style = wx.OK|wx.ICON_INFORMATION
    dlg = wx.MessageDialog(parent, message, title, style)
    ret = dlg.ShowModal()
    dlg.Destroy()
    return ret

def empty_bitmap(width, height, value=255):
    """return empty wx.BitMap"""
    data = array.array('B', [value]*3*width*height)
    return wx.BitmapFromBuffer(width, height, data)

def fix_filename(fname):
    """
    fix string to be a 'good' filename. This may be a more
    restrictive than the OS, but avoids nasty cases.
    """
    out = str(s).translate(TRANS_FILE)
    if out[0] in '-,;[]{}()~`@#':
        out = '_%s' % out
    return out


def FileOpen(parent, message, default_dir=None,
             default_file=None, multiple=False,
             wildcard=None):
    """File Open dialog wrapper.
    returns full path on OK or None on Cancel
    """
    out = None
    if default_dir is None:
        default_dir = os.getcwd()
    if wildcard is None:
        wildcard = 'All files (*.*)|*.*'

    style = wx.FD_OPEN|wx.FD_CHANGE_DIR
    if multiple:
        style = style|wx.MULTIPLE
    dlg = wx.FileDialog(parent, message=message,
                        defaultFile=default_file,
                        defaultDir=default_dir,
                        wildcard=wildcard,
                        style=style)

    out = None
    if dlg.ShowModal() == wx.ID_OK:
        out = os.path.abspath(dlg.GetPath())
    dlg.Destroy()
    return out

def FileSave(parent, message, default_file=None,
             default_dir=None,   wildcard=None):
    "File Save dialog"
    out = None
    if wildcard is None:
        wildcard = 'All files (*.*)|*.*'

    if default_dir is None:
        default_dir = os.getcwd()

    dlg = wx.FileDialog(parent, message=message,
                        defaultFile=default_file,
                        wildcard=wildcard,
                        style=wx.FD_SAVE|wx.FD_CHANGE_DIR)
    if dlg.ShowModal() == wx.ID_OK:
        out = os.path.abspath(dlg.GetPath())
    dlg.Destroy()
    return out


def SelectWorkdir(parent,  message='Select Working Folder...'):
    "prompt for and change into a working directory "
    dlg = wx.DirDialog(parent, message,
                       style=wx.DD_DEFAULT_STYLE|wx.DD_CHANGE_DIR)

    path = os.path.abspath(os.curdir)
    dlg.SetPath(path)
    if  dlg.ShowModal() == wx.ID_CANCEL:
        return None
    path = os.path.abspath(dlg.GetPath())
    dlg.Destroy()
    os.chdir(path)
    return path

class Closure:
    """A very simple callback class to emulate a closure (reference to
    a function with arguments) in python.

    This class holds a user-defined function to be executed when the
    class is invoked as a function.  This is useful in many situations,
    especially for 'callbacks' where lambda's are quite enough.
    Many Tkinter 'actions' can use such callbacks.

    >>>def my_action(x=None):
    ...    print('my action: x = ', x)
    >>>c = Closure(my_action,x=1)
    ..... sometime later ...
    >>>c()
     my action: x = 1
    >>>c(x=2)
     my action: x = 2

    based on Command class from J. Grayson's Tkinter book.
    """
    def __init__(self, func=None, *args, **kws):
        self.func  = func
        self.kws   = kws
        self.args  = args

    def __call__(self,  *args, **kws):
        self.kws.update(kws)
        if callable(self.func):
            self.args = args
            return self.func(*self.args, **self.kws)


class FloatCtrl(wx.TextCtrl):
    """ Numerical Float Control::
    a wx.TextCtrl that allows only numerical input, can take a precision argument
    and optional upper / lower bounds
    Options:

    """
    def __init__(self, parent, value='', minval=None, maxval=None,
                 precision=3, bell_on_invalid = True,
                 act_on_losefocus=False,
                 action=None, action_kw=None, **kws):

        self.__digits = '0123456789.-'
        self.__prec   = precision
        if precision is None:
            self.__prec = 0
        self.format   = '%%.%if' % self.__prec
        self.is_valid = True
        self.__val = set_float(value)
        self.__max = set_float(maxval)
        self.__min = set_float(minval)
        self.__bound_val = None
        self.__mark = None
        self.__action = None

        self.fgcol_valid   = "Black"
        self.bgcol_valid   = "White"
        self.fgcol_invalid = "Red"
        self.bgcol_invalid = (254, 254, 80)
        self.bell_on_invalid = bell_on_invalid
        self.act_on_losefocus = act_on_losefocus

        # set up action
        if action_kw is None:
            action_kw = {}
        self.SetAction(action, **action_kw)

        this_sty =  wx.TE_PROCESS_ENTER|wx.TE_RIGHT
        if 'style' in kws:
            this_sty = this_sty | kws['style']
        kws['style'] = this_sty

        wx.TextCtrl.__init__(self, parent, wx.ID_ANY, **kws)

        self.__CheckValid(self.__val)
        self.SetValue(self.__val)

        self.Bind(wx.EVT_CHAR, self.OnChar)
        self.Bind(wx.EVT_TEXT, self.OnText)

        self.Bind(wx.EVT_SET_FOCUS,  self.OnSetFocus)
        self.Bind(wx.EVT_KILL_FOCUS, self.OnKillFocus)
        self.__GetMark()

    def SetAction(self, action, **kws):
        "set callback action"
        if callable(action):
            self.__action = Closure(action, **kws)

    def SetPrecision(self, prec=0):
        "set precision"
        self.__prec = prec
        self.format = '%%.%if' % prec

    def __GetMark(self):
        " keep track of cursor position within text"
        try:
            self.__mark = min(wx.TextCtrl.GetSelection(self)[0],
                              len(wx.TextCtrl.GetValue(self).strip()))
        except:
            self.__mark = 0

    def __SetMark(self, mark=None):
        "set mark for later"
        if mark is None:
            mark = self.__mark
        self.SetSelection(mark, mark)

    def SetValue(self, value=None, act=True):
        " main method to set value "
        if value is None:
            value = wx.TextCtrl.GetValue(self).strip()
        self.__CheckValid(value)
        self.__GetMark()
        value = set_float(value)
        if value is not None:
            wx.TextCtrl.SetValue(self, self.format % value)

        if self.is_valid and callable(self.__action) and act:
            self.__action(value=self.__val)
        elif not self.is_valid and self.bell_on_invalid:
            wx.Bell()

        self.__SetMark()

    def OnKillFocus(self, event):
        "focus lost"
        self.__GetMark()
        if self.act_on_losefocus and callable(self.__action):
            self.__action(value=self.__val)
        event.Skip()

    def OnSetFocus(self, event):
        "focus gained - resume editing from last mark point"
        self.__SetMark()
        event.Skip()

    def OnChar(self, event):
        """ on Character event"""
        key   = event.GetKeyCode()
        entry = wx.TextCtrl.GetValue(self).strip()
        pos   = wx.TextCtrl.GetSelection(self)
        # really, the order here is important:
        # 1. return sends to ValidateEntry
        if key == wx.WXK_RETURN:
            if not self.is_valid:
                wx.TextCtrl.SetValue(self, self.format % set_float(self.__bound_val))
            else:
                self.SetValue(entry)
            return

        # 2. other non-text characters are passed without change
        if (key < wx.WXK_SPACE or key == wx.WXK_DELETE or key > 255):
            event.Skip()
            return

        # 3. check for multiple '.' and out of place '-' signs and ignore these
        #    note that chr(key) will now work due to return at #2

        has_minus = '-' in entry
        ckey = chr(key)
        if ((ckey == '.' and (self.__prec == 0 or '.' in entry) ) or
            (ckey == '-' and (has_minus or  pos[0] != 0)) or
            (ckey != '-' and  has_minus and pos[0] == 0)):
            return
        # 4. allow digits, but not other characters
        if chr(key) in self.__digits:
            event.Skip()


    def OnText(self, event=None):
        "text event"
        try:
            if event.GetString() != '':
                self.__CheckValid(event.GetString())
        except:
            pass
        event.Skip()

    def GetValue(self):
        if self.__prec > 0:
            return set_float("%%.%ig" % (self.__prec) % (self.__val))
        else:
            return int(self.__val)

    def GetMin(self):
        "return min value"
        return self.__min

    def GetMax(self):
        "return max value"
        return self.__max

    def SetMin(self, val):
        "set min value"
        self.__min = set_float(val)

    def SetMax(self, val):
        "set max value"
        self.__max = set_float(val)

    def __CheckValid(self, value):
        "check for validity of value"
        val = self.__val
        self.is_valid = True
        try:
            val = set_float(value)
            if self.__min is not None and (val < self.__min):
                self.is_valid = False
                val = self.__min
            if self.__max is not None and (val > self.__max):
                self.is_valid = False
                val = self.__max
        except:
            self.is_valid = False
        self.__bound_val = self.__val = val
        fgcol, bgcol = self.fgcol_valid, self.bgcol_valid
        if not self.is_valid:
            fgcol, bgcol = self.fgcol_invalid, self.bgcol_invalid

        self.SetForegroundColour(fgcol)
        self.SetBackgroundColour(bgcol)
        self.Refresh()


class NumericCombo(wx.ComboBox):
    """
    Numeric Combo: ComboBox with numeric-only choices
    """
    def __init__(self, parent, choices, precision=3,
                 init=0, width=80):

        self.fmt = "%%.%if" % precision
        self.choices  = choices
        schoices = [self.fmt % i for i in self.choices]
        wx.ComboBox.__init__(self, parent, -1, '', (-1, -1), (width, -1),
                             schoices, wx.CB_DROPDOWN|wx.TE_PROCESS_ENTER)

        init = min(init, len(self.choices))
        self.SetStringSelection(schoices[init])
        self.Bind(wx.EVT_TEXT_ENTER, self.OnEnter)

    def OnEnter(self, event=None):
        "text enter event handler"
        thisval = float(event.GetString())

        if thisval not in self.choices:
            self.choices.append(thisval)
            self.choices.sort()

        self.Clear()
        self.AppendItems([self.fmt % i for i in self.choices])
        self.SetSelection(self.choices.index(thisval))

class SimpleText(wx.StaticText):
    "simple static text wrapper"
    def __init__(self, parent, label, minsize=None,
                 font=None, colour=None, bgcolour=None,  **kws):

        wx.StaticText.__init__(self, parent, -1,
                               label=label, **kws)

        if minsize is not None:
            self.SetMinSize(minsize)
        if font is not None:
            self.SetFont(font)
        if colour is not None:
            self.SetForegroundColour(colour)
        if bgcolour is not None:
            self.SetBackgroundColour(colour)

class HyperText(wx.StaticText):
    """HyperText is a simple extension of wx.StaticText that

       1. adds an underscore to the lable to appear to be a hyperlink
       2. performs the supplied action on Left-Up button events
    """
    def  __init__(self, parent, label, action=None, colour=(50, 50, 180)):
        self.action = action
        wx.StaticText.__init__(self, parent, -1, label=label)
        font  = self.GetFont() # .Bold()
        font.SetUnderlined(True)
        self.SetFont(font)
        self.SetForegroundColour(colour)
        self.Bind(wx.EVT_LEFT_UP, self.OnSelect)

    def OnSelect(self, evt=None):
        "Left-Up Event Handler"
        if self.action is not None:
            self.action(evt=evt, label=self.GetLabel())
        evt.Skip()

class DateTimeCtrl(object):
    """
    Simple Combined date/time control
    """
    def __init__(self, parent, name='datetimectrl', use_now=False):
        self.name = name
        panel = self.panel = wx.Panel(parent)
        bgcol = wx.Colour(250, 250, 250)

        datestyle = wx.DP_DROPDOWN|wx.DP_SHOWCENTURY|wx.DP_ALLOWNONE

        self.datectrl = wx.DatePickerCtrl(panel, size=(120, -1),
                                          style=datestyle)
        self.timectrl = masked.TimeCtrl(panel, -1, name=name,
                                        limited=False,
                                        fmt24hr=True, oob_color=bgcol)
        timerheight = self.timectrl.GetSize().height
        spinner = wx.SpinButton(panel, -1, wx.DefaultPosition,
                                (-1, timerheight), wx.SP_VERTICAL )
        self.timectrl.BindSpinButton(spinner)
        sizer = wx.BoxSizer(wx.HORIZONTAL)
        sizer.Add(self.datectrl, 0, wx.ALIGN_CENTER)
        sizer.Add(self.timectrl, 0, wx.ALIGN_CENTER)
        sizer.Add(spinner, 0, wx.ALIGN_LEFT)
        panel.SetSizer(sizer)
        sizer.Fit(panel)
        if use_now:
            self.timectrl.SetValue(wx.DateTime_Now())



#!/usr/bin/env python
#  M Newville <newville@cars.uchicago.edu>
#  The University of Chicago, 2010
#  Epics Open License
#
# Epics Database Records (DBR) Constants and Definitions
#  most of the code here is copied from db_access.h
#
""" constants and declaration of data types for Epics database records
This is mostly copied from CA header files
"""
import ctypes
import functools
import os
import platform
import enum

HAS_NUMPY = False
try:
    import numpy
    HAS_NUMPY = True
except ImportError:
    pass

PY64_WINDOWS = (os.name == 'nt' and platform.architecture()[0].startswith('64'))

# EPICS Constants
ECA_NORMAL = 1
ECA_TIMEOUT = 80
ECA_IODONE = 339
ECA_ISATTACHED = 424
ECA_BADCHID = 410

CS_CONN    = 2
OP_CONN_UP = 6
OP_CONN_DOWN = 7

CS_NEVER_SEARCH = 4
#
# Note that DBR_XXX should be replaced with dbr.XXX
#
STRING = 0
INT    = 1
SHORT  = 1
FLOAT  = 2
ENUM   = 3
CHAR   = 4
LONG   = 5
DOUBLE = 6

TIME_STRING  = 14
TIME_INT     = 15
TIME_SHORT   = 15
TIME_FLOAT   = 16
TIME_ENUM    = 17
TIME_CHAR    = 18
TIME_LONG    = 19
TIME_DOUBLE  = 20

CTRL_STRING  = 28
CTRL_INT     = 29
CTRL_SHORT   = 29
CTRL_FLOAT   = 30
CTRL_ENUM    = 31
CTRL_CHAR    = 32
CTRL_LONG    = 33
CTRL_DOUBLE  = 34

PUT_ACKT     = 35
PUT_ACKS     = 36
STSACK_STRING = 37
CLASS_NAME   = 38

MAX_STRING_SIZE      = 40
MAX_UNITS_SIZE       =  8
MAX_ENUM_STRING_SIZE = 26
MAX_ENUMS            = 16

#EPICS2UNIX_EPOCH = 631173600.0 - time.timezone
EPICS2UNIX_EPOCH = 631152000.0

# create_subscription mask constants
DBE_VALUE = 1
DBE_LOG = 2
DBE_ALARM = 4
DBE_PROPERTY = 8

chid_t   = ctypes.c_long

# Note that Windows needs to be told that chid is 8 bytes for 64-bit
if PY64_WINDOWS:
    chid_t = ctypes.c_int64

short_t  = ctypes.c_short
ushort_t = ctypes.c_ushort
int_t    = ctypes.c_int
uint_t   = ctypes.c_uint
long_t   = ctypes.c_long
ulong_t  = ctypes.c_ulong
float_t  = ctypes.c_float
double_t = ctypes.c_double
byte_t   = ctypes.c_byte
ubyte_t  = ctypes.c_ubyte
string_t = ctypes.c_char * MAX_STRING_SIZE
char_t   = ctypes.c_char
char_p   = ctypes.c_char_p
void_p   = ctypes.c_void_p
py_obj   = ctypes.py_object


value_offset = [39*ctypes.c_short]

# extended DBR types:
class TimeStamp(ctypes.Structure):
    "emulate epics timestamp"
    _fields_ = [('secs', uint_t), ('nsec', uint_t)]

_STAT_SEV    = (('status', short_t), ('severity', short_t))
_STAT_SEV_TS = (('status', short_t), ('severity', short_t),
                ('stamp', TimeStamp))
_UNITS       = ('units', char_t * MAX_UNITS_SIZE)

def make_unixtime(stamp):
    "UNIX timestamp (seconds) from Epics TimeStamp structure"
    return EPICS2UNIX_EPOCH + stamp.secs + 1.e-6*int(1.e-3*stamp.nsec)


class time_string(ctypes.Structure):
    "dbr time string"
    _fields_ = list(_STAT_SEV_TS) + [('value', MAX_STRING_SIZE*char_t)]


class time_short(ctypes.Structure):
    "dbr time short"
    _fields_ = list(_STAT_SEV_TS) + [('RISC_pad',  short_t),
                                     ('value',     short_t)]

class time_float(ctypes.Structure):
    "dbr time float"
    _fields_ = list(_STAT_SEV_TS) + [('value',  float_t)]

class time_enum(ctypes.Structure):
    "dbr time enum"
    _fields_ = list(_STAT_SEV_TS) + [('RISC_pad',  short_t),
                                     ('value',    ushort_t)]

class time_char(ctypes.Structure):
    "dbr time char"
    _fields_ = list(_STAT_SEV_TS) + [('RISC_pad0', short_t),
                                     ('RISC_pad1', byte_t),
                                     ('value',     byte_t)]

class time_long(ctypes.Structure):
    "dbr time long"
    _fields_ = list(_STAT_SEV_TS) + [('value', int_t)]


class time_double(ctypes.Structure):
    "dbr time double"
    _fields_ = list(_STAT_SEV_TS) + [('RISC_pad', int_t),
                                     ('value',    double_t)]

# DBR types with full control and graphical fields
# yes, this strange order is as in db_access.h!!!
ctrl_limits = ('upper_disp_limit',   'lower_disp_limit',
               'upper_alarm_limit',  'upper_warning_limit',
               'lower_warning_limit','lower_alarm_limit',
               'upper_ctrl_limit',   'lower_ctrl_limit')

def _gen_ctrl_lims(t=short_t):
    "create types for control limits"
    return  [(s, t) for s in  ctrl_limits]

class ctrl_enum(ctypes.Structure):
    "dbr ctrl enum"
    _fields_ = list(_STAT_SEV)
    _fields_.extend([ ('no_str', short_t),
                      ('strs', (char_t * MAX_ENUM_STRING_SIZE) * MAX_ENUMS),
                      ('value',    ushort_t)])

class ctrl_short(ctypes.Structure):
    "dbr ctrl short"
    _fields_ = list(_STAT_SEV) + [_UNITS] +  _gen_ctrl_lims(t=short_t)
    _fields_.extend([('value', short_t )])

class ctrl_char(ctypes.Structure):
    "dbr ctrl long"
    _fields_ = list(_STAT_SEV) +[_UNITS] +  _gen_ctrl_lims(t=byte_t)
    _fields_.extend([('RISC_pad', byte_t), ('value', ubyte_t)])

class ctrl_long(ctypes.Structure):
    "dbr ctrl long"
    _fields_ = list(_STAT_SEV) +[_UNITS] +  _gen_ctrl_lims(t=int_t)
    _fields_.extend([('value', int_t)])

class ctrl_float(ctypes.Structure):
    "dbr ctrl float"
    _fields_ = list(_STAT_SEV)
    _fields_.extend([('precision',   short_t),
                     ('RISC_pad',    short_t)] + [_UNITS])
    _fields_.extend( _gen_ctrl_lims(t=float_t) )
    _fields_.extend([('value', float_t)])


class ctrl_double(ctypes.Structure):
    "dbr ctrl double"
    _fields_ = list(_STAT_SEV)
    _fields_.extend([('precision',   short_t),
                     ('RISC_pad',    short_t)] + [_UNITS])
    _fields_.extend( _gen_ctrl_lims(t=double_t) )
    _fields_.extend([('value',       double_t)])


NP_Map = {}
if HAS_NUMPY:
    NP_Map = {INT:    numpy.int16,
              FLOAT:  numpy.float32,
              ENUM:   numpy.uint16,
              CHAR:   numpy.uint8,
              LONG:   numpy.int32,
              DOUBLE: numpy.float64}


# map of Epics DBR types to ctypes types
Map = {STRING: string_t,
       INT:    short_t,
       FLOAT:  float_t,
       ENUM:   ushort_t,
       CHAR:   ubyte_t,
       LONG:   int_t,
       DOUBLE: double_t,

       TIME_STRING: time_string,
       TIME_INT: time_short,
       TIME_SHORT:  time_short,
       TIME_FLOAT: time_float,
       TIME_ENUM:  time_enum,
       TIME_CHAR:  time_char,
       TIME_LONG:  time_long,
       TIME_DOUBLE: time_double,
       # Note: there is no ctrl string in the C definition
       CTRL_STRING:   time_string,
       CTRL_SHORT: ctrl_short,
       CTRL_INT:   ctrl_short,
       CTRL_FLOAT: ctrl_float,
       CTRL_ENUM:  ctrl_enum,
       CTRL_CHAR:  ctrl_char,
       CTRL_LONG:  ctrl_long,
       CTRL_DOUBLE: ctrl_double
       }

def native_type(ftype):
    "return native field type from TIME or CTRL variant"
    if ftype == CTRL_STRING:
        ftype = TIME_STRING
    ntype = ftype
    if ftype > CTRL_STRING:
        ntype -= CTRL_STRING
    elif ftype >= TIME_STRING:
        ntype -= TIME_STRING
    return ntype

def Name(ftype, reverse=False):
    """ convert integer data type to dbr Name, or optionally reverse that
    look up (that is, name to integer)"""
    m = {STRING: 'STRING',
         INT: 'INT',
         FLOAT: 'FLOAT',
         ENUM: 'ENUM',
         CHAR: 'CHAR',
         LONG: 'LONG',
         DOUBLE: 'DOUBLE',

         TIME_STRING: 'TIME_STRING',
         TIME_SHORT: 'TIME_SHORT',
         TIME_FLOAT: 'TIME_FLOAT',
         TIME_ENUM: 'TIME_ENUM',
         TIME_CHAR: 'TIME_CHAR',
         TIME_LONG: 'TIME_LONG',
         TIME_DOUBLE: 'TIME_DOUBLE',

         CTRL_STRING: 'CTRL_STRING',
         CTRL_SHORT: 'CTRL_SHORT',
         CTRL_FLOAT: 'CTRL_FLOAT',
         CTRL_ENUM: 'CTRL_ENUM',
         CTRL_CHAR: 'CTRL_CHAR',
         CTRL_LONG: 'CTRL_LONG',
         CTRL_DOUBLE: 'CTRL_DOUBLE',
         }
    if reverse:
        name = ftype.upper()
        if name in list(m.values()):
            for key, val in m.items():
                if name == val: return key
    return m.get(ftype, 'unknown')

def cast_args(args):
    """returns casted array contents
    returns: [dbr_ctrl or dbr_time struct,
              count * native_type structs]

    If data is already of a native_type, the first
    value in the list will be None.

    If type is not known, both the first and the second
    value in the list will be None.
    """
    ftype = args.type
    if ftype not in Map:
        # Type was not found in the map.
        # As a consequence, the code cannot
        # interpret the pointer properly without
        # a risk of accessing invalid memory area.
        return [None, None]

    ntype = native_type(ftype)
    if ftype != ntype:
        native_start = args.raw_dbr + value_offset[ftype]
        return [ctypes.cast(args.raw_dbr,
                            ctypes.POINTER(Map[ftype])).contents,
                ctypes.cast(native_start,
                            ctypes.POINTER(args.count * Map[ntype])).contents]
    return [None, ctypes.cast(args.raw_dbr,
                          ctypes.POINTER(args.count * Map[ftype])).contents]


if PY64_WINDOWS:
    def make_callback(func, args):
        """ make callback function"""
        # note that ctypes.POINTER is needed for 64-bit Python on Windows
        @functools.wraps(func)
        def wrapped(arg, **kwargs):
            # On 64-bit Windows, `arg.contents` seems to be equivalent to other
            # platforms' `arg`
            if hasattr(arg, 'contents'):
                return func(arg.contents, **kwargs)
            return func(arg, **kwargs)

        return ctypes.CFUNCTYPE(None, ctypes.POINTER(args))(wrapped)
else:
    def make_callback(func, args):
        """ make callback function"""
        return ctypes.CFUNCTYPE(None, args)(func)


class event_handler_args(ctypes.Structure):
    "event handler arguments"
    _fields_ = [('usr',     ctypes.py_object),
                ('chid',    chid_t),
                ('type',    long_t),
                ('count',   long_t),
                ('raw_dbr', void_p),
                ('status',  int_t)]

class connection_args(ctypes.Structure):
    "connection arguments"
    _fields_ = [('chid', chid_t),
                ('op', long_t)]


class access_rights_handler_args(ctypes.Structure):
    "access rights arguments"
    _fields_ = [('chid', chid_t),
                ('access', ubyte_t)]


class DefaultIntEnum(enum.IntEnum):

    @classmethod
    def _missing_(cls, value):
        unknown = int.__new__(cls)
        unknown._name_ = "UNKNOWN"
        unknown._value_ = value
        return unknown

    def __repr__(self):
        return self.name


class AlarmStatus(DefaultIntEnum):
    NO_ALARM = 0
    READ = 1
    WRITE = 2
    HIHI = 3
    HIGH = 4
    LOLO = 5
    LOW = 6
    STATE = 7
    COS = 8
    COMM = 9
    TIMEOUT = 10
    HW_LIMIT = 11
    CALC = 12
    SCAN = 13
    LINK = 14
    SOFT = 15
    BAD_SUB = 16
    UDF = 17
    DISABLE = 18
    SIMM = 19
    READ_ACCESS = 20
    WRITE_ACCESS = 21

class AlarmSeverity(DefaultIntEnum):
    NO_ALARM = 0
    MINOR = 1
    MAJOR = 2
    INVALID = 3



from epics.utils import IOENCODING
import epics.autosave
import time
epics.autosave.save_pvs('AutoSaveTest.req', 'tmp.sav')

time.sleep(0.5)
f = open('tmp.sav','r', encoding=IOENCODING)

data  = f.read()

if len(data) > 50:
    print("AutoSave worked... data written to file 'tmp.sav'")



#!/usr/bin/env python

"""ca_util.py is a wrapper around pyepics that allows the caller to write,
e.g.,
    caget("xxx:m1")
instead of having to write
    m1 = CaChannel()
    m1.searchw("xxx:m1")
    m1.getw()
Also, ca_util defends against null PV names and some effects of short-term
CA disconnections, and it can verify that caput*() operations succeeded.

This is a port of Tim Mooney's wrapper around CaChannel to pyepics.
"""

# Port of Tim Mooney's wrapper around CaChannel to pyepics
# Tim Mooney 12/05/2008

code = '''
version = "3.0"

import epics
import time
import sys

# DBR types
# ca.DBR_STRING = 0
# ca.DBR_SHORT = 1 
# ca.DBR_INT = 1 
# ca.DBR_FLOAT = 2
# ca.DBR_ENUM = 3
# ca.DBR_CHAR = 4
# ca.DBR_LONG = 5
# ca.DBR_DOUBLE = 6

#######################################################################
# Human readable exception description
#	try:
#		x = x + 1
#	except:
#		print formatExceptionInfo()
import sys
import traceback
def formatExceptionInfo(maxTBlevel=5):
	cla, exc, trbk = sys.exc_info()
	excName = cla.__name__
	try:
		excArgs = exc.__dict__["args"]
	except KeyError:
		excArgs = "<no args>"
	excTb = traceback.format_tb(trbk, maxTBlevel)
	return (excName, excArgs, excTb)

#######################################################################
# channel-access connection states
ca_states = {}
# ...from cadef.h:
ca_states[ca.cs_never_conn] = "never connected"
ca_states[ca.cs_prev_conn] = "previously connected"
ca_states[ca.cs_conn] = "connected"
ca_states[ca.cs_closed] = "closed"
# ...from CaChannel.py (only with the patch from 04/29/04):
try:
	ca_states[ca.cs_never_search] = "never searched"
except AttributeError:
	file = getCaChannelFileName()
	using_old_CaChannel = 1
	print "ca_util: You're using an old version of CaChannel (%s)." % getCaChannelFileName()
	pass
	


#######################################################################
# default settings for ca_util
defaultTimeout = None		# 'None' means use CaChannel's timeout
defaultRetries = 3
readCheckTolerance = None	# 'None" means don't check

def set_ca_util_defaults(timeout=None, retries=None, read_check_tolerance=None):
	"""
	usage: old = set_ca_util_defaults(timeout=None, retries=None,
	             read_check_tolerance=None)
	alternate: set_ca_util_defaults(defaultsList), where defaultsList is like
	the list returned by get_ca_util_defaults()
	Setting an argument to the string "NONE" disables it.
	Returns the list of previous default values:
	    [defaultTimeout, defaultRetries, readCheckTolerance]
	"""
	global defaultTimeout, defaultRetries, readCheckTolerance
	old = [defaultTimeout, defaultRetries, readCheckTolerance]
	if type(timeout) == type([]):
		argList = timeout
		timeout = argList[0]
		retries = argList[1]
		read_check_tolerance = argList[2]
	if (timeout!=None) : defaultTimeout = timeout
	if (retries!=None) : defaultRetries = retries
	if (read_check_tolerance!=None) : readCheckTolerance = read_check_tolerance
	return old

def get_ca_util_defaults():
	"""
	usage: myList = get_ca_util_defaults()
	myList is set to [defaultTimeout, defaultRetries, readCheckTolerance]
	"""
	global defaultTimeout, defaultRetries, readCheckTolerance
	return [defaultTimeout, defaultRetries, readCheckTolerance]

def set_ca_util_default_timeout(timeout=None):
	"""
	usage: old = set_ca_util_default_timeout(timeout=None)
	If timeout == "NONE", then ca_util doesn't specify any timeout in
	calls to underlying software.
	Returns previous default timeout.
	"""
	global defaultTimeout
	old = defaultTimeout
	defaultTimeout = timeout
	return old

def get_ca_util_default_timeout():
	global defaultTimeout
	return defaultTimeout

def set_ca_util_default_retries(retries=None):
	"""
	usage: old = set_ca_util_default_retries(retries=None)
	If retries == "NONE", then ca_util doesn't do any retries.
	Returns previous default retries.
	"""
	global defaultRetries
	old = defaultRetries
	defaultRetries = retries
	return old

def get_ca_util_default_retries():
	global defaultRetries
	return defaultRetries

def set_ca_util_default_read_check_tolerance(read_check_tolerance=None):
	"""
	usage: old = set_ca_util_default_read_check_tolerance(read_check_tolerance=None)
	If read_check_tolerance == "NONE", then ca_util doesn't compare the value
	it reads to the value it wrote.
	Returns previous default tolerance.
	"""
	global readCheckTolerance
	old = readCheckTolerance
	readCheckTolerance = read_check_tolerance
	return old

def get_ca_util_default_read_check_tolerance():
	global readCheckTolerance
	return readCheckTolerance


#######################################################################
# The dictionary, cadict, will be used to associate PV names with the
# machinery required to talk to EPICS PV's.  If no entry is found (the
# name hasn't been used yet in a ca call), then we create a new instance
# of CaChannel, connect it to the PV, and put it in the dictionary.  We also
# include a flag some of the ca_util routines can use to check if a callback
# has occurred for this PV.

class cadictEntry:
	def __init__(self, channel):
		self.channel = channel
		self.callbackReceived = 0	# reserved for use by caputw()
		self.field_type = channel.field_type()
		self.element_count = channel.element_count()
		#self.host_name = channel.host_name()

cadict = {}

#######################################################################
ca_utilExceptionStrings = ["No name was provided.", "Readback disagrees with put value.",
                           "PV is not connected."]
EXCEPTION_NULL_NAME          = 0
EXCEPTION_READBACK_DISAGREES = 1
EXCEPTION_NOT_CONNECTED      = 2

class ca_utilException(Exception):
	def __init__(self, *args):
		Exception.__init__(self, *args)
		self.errorNumber = args[0]

	def __int__(self):
		return int(self.errorNumber)

	def __str__(self):
		return ca_utilExceptionStrings[self.errorNumber]


#######################################################################
def convertToType(type, value):
	if type == ca.DBR_STRING:
		return str(value)
	elif type == ca.DBR_SHORT or type == ca.DBR_INT or type == ca.DBR_LONG:
		try:
			n = int(value)
		except:
			n = 0
		return n
	elif type == ca.DBR_FLOAT or type == ca.DBR_DOUBLE:
		try:
			n = float(value)
		except:
			n = 0.0
		return n
	elif type == ca.DBR_ENUM:
		return value
	elif type == ca.DBR_CHAR:
		return value
	else:
		return value

#######################################################################
def checkName(name, timeout=None, retries=None):
	"""
	usage: checkName("xxx:m1.VAL", timeout=None, retries=None)
	Intended for internal use by ca_util functions.
	"""

	global cadict, defaultTimeout, defaultRetries
	if not name:
		raise ca_utilException, EXCEPTION_NULL_NAME
		return

	if ((timeout == None) and (defaultTimeout != None)): timeout = defaultTimeout
	if (timeout == "NONE"): timeout = None

	if ((retries == None) and (defaultRetries != None)): retries = defaultRetries
	if ((retries == None) or (retries == "NONE")): retries = 0

	tries = 0
	while (not cadict.has_key(name)) and (tries <= retries):
		# Make a new entry in the PV-name dictionary
		try:
			channel = CaChannel.CaChannel()
			if (timeout != None): channel.setTimeout(timeout)
			channel.searchw(name)
			cadict[name] = cadictEntry(channel)
		except CaChannel.CaChannelException, status:
			del channel
			tries += 1

	if (not cadict.has_key(name)):
		print "ca_util.checkName: Can't connect to '%s'" % name
		raise CaChannel.CaChannelException, status

#######################################################################
def castate(name=None, timeout=None, retries=None):
	"""usage: val = castate("xxx:m1.VAL", timeout=None, retries=None)
	   Try to read a PV, to find out whether it's really connected, and
	   whether caller is permitted to read and write it, without allowing
	   any exceptions to be thrown at the caller.
	"""

	global cadict, defaultTimeout, defaultRetries

	if not name: return "Null name has no state"

	# The only reliable way to check the *current* state of a PV is to attempt to use it.
	try:
		val = caget(name, timeout=timeout, retries=retries)
	except CaChannel.CaChannelException, status:
		pass

	try:
		checkName(name, timeout=timeout)
	except CaChannel.CaChannelException, status:
		return "not connected"
	except:
		return "error"

	try:
		state = cadict[name].channel.state()
	except CaChannel.CaChannelException, status:
		return "not connected"
	except:
		return "error"
	else:
		try:
			read_access = cadict[name].channel.read_access()
			write_access = cadict[name].channel.write_access()
			if ca_states.has_key(state):
				s = ca_states[state]
			else:
				s = "unknown state"
			if not read_access: s += ", noread"
			if not write_access: s += ", nowrite"
			return s
		except:
			return "error"

#######################################################################
def caget(name, timeout=None, retries=None, req_type=None, req_count=None):
	"""usage: val = caget("xxx:m1.VAL", timeout=None, retries=None,
	                req_type=None, req_count=None)"""

	global cadict, defaultTimeout, defaultRetries

	if not name:
		print "caget: no PV name supplied"
		raise ca_utilException, EXCEPTION_NULL_NAME
		return 0
	if ((timeout==None) and (defaultTimeout != None)): timeout = defaultTimeout
	if (timeout == "NONE"): timeout = None
	if ((retries==None) and (defaultRetries != None)): retries = defaultRetries
	if ((retries == None) or (retries == "NONE")): retries = 0
	retries = max(retries,0)
	retry = retries + 1
	success = 0

	# CaChannel sometimes chokes when it tries to process a channel that has been disconnected.
	# The simplest fix is to clear the channel and reconnect to the PV, which we can do cleanly
	# by deleting our dict entry for the channel, and calling checkName() to make a new entry.

	while ((not success) and (retry > 0)):
		checked = 0
		while ((not checked) and (retry > 0)):
			retry -= 1
			try:
				checkName(name, timeout=timeout)
			except CaChannel.CaChannelException, status:
				if retry <= 0:
					raise CaChannel.CaChannelException, status
					return 0
			else:
				checked = 1

		entry = cadict[name]
		if (timeout != None): entry.channel.setTimeout(timeout)
		if req_type == None:
			req_type=entry.field_type
		# kludge for broken DBR_CHAR
		if req_type == ca.DBR_CHAR:
			req_type = ca.DBR_INT
		if req_count == None:
			req_count = entry.element_count
		req_count = max(0, min(req_count, entry.element_count))
		try:
			if (using_old_CaChannel):
				val = entry.channel.getw(req_type=req_type)
			else:
				val = entry.channel.getw(req_type=req_type, count=req_count)
		except CaChannel.CaChannelException, status:
			#print "getw threw an exception (%s)" % status
			if ((int(status) == ca.ECA_BADTYPE) or (int(status) == ca.ECA_DISCONN)):
				# Delete dictionary entry.  This clears the CA connection.
				print "caget: Repairing CA connection to ", name
				del cadict[name]
				retry += 1
			if retry <= 0:
				raise CaChannel.CaChannelException, status
				return 0
		else:
			success = 1
	return val

def isNumber(s):
	try:
		n = int(s)
	except:
		return False
	return True

#######################################################################
def same(value, readback, native_readback, field_type, read_check_tolerance):
	"""For internal use by ca_util"""
	#print "ca_util.same(): field_type=%s" % field_type
	#print "ca_util.same(): value='%s'; readback='%s', native_readback='%s'" % (str(value), str(readback), str(native_readback))
	#print "ca_util.same(): type(value)=%s; type(readback)=%s, type(native_readback)=%s" % (type(value),
	#	type(readback), type(native_readback))

	if field_type in [ca.DBR_FLOAT, ca.DBR_DOUBLE]:
		return (abs(float(readback)-float(value)) < read_check_tolerance)
	elif field_type in [ca.DBR_INT, ca.DBR_SHORT, ca.DBR_LONG]:
		return (abs(int(readback)-int(value)) == 0)
	elif field_type == ca.DBR_ENUM:
		if str(value) == str(readback):
			return True
		if str(value) == str(native_readback):
			return True
		return False
	else:
		return (str(value) == str(readback))

#######################################################################
def caput(name, value, timeout=None, req_type=None, retries=None, read_check_tolerance=None):
	"""
	usage: caput("xxx:m1.VAL", new_value, timeout=None, req_type=None,
	       retries=None, read_check_tolerance=None)
	Put a value, and optionally check that the value arrived safely.
	If read_check_tolerance == None (or is not supplied) then the default
	read-check tolerance is used.  If read_check_tolerance == "NONE", then no
	read check is done.
	If read_check_tolerance != "NONE", then floating point numbers must be
	closer than the tolerance, and other types must agree exactly.
	Note that defaults for timeout, retries, and read_check_tolerance can be
	set for all ca_util functions, using the command set_ca_util_defaults().
	"""

	_caput("caput", name, value, 0, timeout, req_type, retries, read_check_tolerance)


#######################################################################
def __ca_util_waitCB(epics_args, user_args):
	"""Function for internal use by caputw()."""
	#print "__ca_util_waitCB: %s done\n" % user_args[0]
	cadict[user_args[0]].callbackReceived = 1

#######################################################################
def caputw(name, value, wait_timeout=None, timeout=None, req_type=None, retries=None,
 read_check_tolerance=None):
	"""
	usage: caputw("xxx:m1.VAL", new_value, wait_timeout=None, timeout=None,
	       req_type=None, retries=None, read_check_tolerance=None)
	Put a value, optionally check that the value arrived safely, and wait (no
	longer than wait_timeout) for processing to complete. If
	read_check_tolerance == None (or is not supplied) then the default
	read-check tolerance is used.  If read_check_tolerance == "NONE", then no
	read check is done. If read_check_tolerance != "NONE", then floating point
	numbers must be closer than the tolerance, and other types must agree
	exactly. Note that defaults for timeout, retries, and read_check_tolerance
	can be set for all ca_util functions, using the command
	set_ca_util_defaults().
	"""

	_caput("caputw", name, value, wait_timeout, timeout, req_type, retries, read_check_tolerance)


#######################################################################
def _caput(function, name, value, wait_timeout=None, timeout=None, req_type=None, retries=None, read_check_tolerance=None):

	global cadict, defaultTimeout, defaultRetries, readCheckTolerance

	#print function
	if not name:
		print "%s: no PV name supplied" % function
		raise ca_utilException, EXCEPTION_NULL_NAME
		return
	if ((timeout == None) and (defaultTimeout != None)): timeout = defaultTimeout
	if ((retries == None) and (defaultRetries != None)): retries = defaultRetries
	if ((retries == None) or (retries == "NONE")): retries = 0
	if ((read_check_tolerance == None) and (readCheckTolerance != None)):
		read_check_tolerance = readCheckTolerance

	retries = max(retries,0)
	retry = retries + 1
	success = 0

	checkName(name, timeout=timeout, retries=retries)

	while ((not success) and (retry > 0)):

		retry -= 1
		entry = cadict[name]

		state = castate(name, timeout)
		#print "%s: state='%s'" % (function, state)
		if (state != 'connected'):
			print "%s: Repairing CA connection to '%s'" % (function, name)
			del cadict[name]
			retry += 1
		else:
			if req_type == None:
				req_type=entry.field_type
			if ((timeout != None) and (timeout != "NONE")): entry.channel.setTimeout(timeout)
			entry.callbackReceived = 0 # in case we're doing caputw()
			#value = convertToType(value, req_type)
			try:
				if function == "caput":
					entry.channel.putw(value, req_type=req_type)
				else: #caputw
					retval = entry.channel.array_put_callback(value,req_type,entry.element_count,__ca_util_waitCB,name)
			except CaChannel.CaChannelException, status:
				print "put() threw an exception (%s)" % status
				if ((int(status) == ca.ECA_BADTYPE) or (int(status) == ca.ECA_DISCONN)):
					# Delete dictionary entry.  This clears the CA connection.
					print "%s: Repairing CA connection to '%s'" % (function, name)
					del cadict[name]
					retry += 1
				if retry <= 0:
					raise CaChannel.CaChannelException, status
					entry.callbackReceived = 1
					return
			else:
				if ((read_check_tolerance == None) or (read_check_tolerance == "NONE")):
					success = True
				else:
					if timeout:
						ca.pend_io(timeout)
					else:
						ca.pend_io(1.0)
					readback_success = False
					count = 0
					while ((not readback_success) and (count < retries+1)):
						try:
							readback = caget(name, req_type=req_type)
							native_readback = caget(name)
							readback_success = True
							if same(value, readback, native_readback, entry.field_type, read_check_tolerance):
								success = True
								#print "%s: Success\n" % (function)
							else:
								print "%s: readback '%s' disagrees with the value '%s' we wrote." % (function, readback, value)
								raise ca_utilException, EXCEPTION_READBACK_DISAGREES
								entry.callbackReceived = 1
						except CaChannel.CaChannelException, status:
							print "%s: exception during readback." % (function)
							count += 1

	if success and (function == "caputw"):
		start_time = time.time()
		timed_out = 0
		while (not entry.callbackReceived) and (not timed_out):
			#print "waiting for ", name
			time.sleep(0.1)
			#ca.pend_io(0.1)
			ca.poll()
			if (not wait_timeout):
				timed_out = 0
			else:
				timed_out = ((time.time()-start_time) > wait_timeout)

		if not entry.callbackReceived:
			print "Execution not completed by wait_timeout (%d seconds)" % wait_timeout

#######################################################################
def camonitor(name, function, user_args=None, timeout=None, retries=None):
	"""
	usage: camonitor("xxx:m1.VAL", python_function, user_args, timeout=None,
	       retries=None)
	Don't forget to call ca.pend_event(<pend_time_in_seconds>) periodically.
	"""

	global defaultTimeout, defaultRetries

	if not name:
		print "camonitor: no PV name supplied"
		raise ca_utilException, EXCEPTION_NULL_NAME
		return
	if not function:
		print "camonitor: no callback function supplied"
		raise ca_utilException, EXCEPTION_NULL_NAME
		return
	if not user_args: user_args = name
	if ((timeout==None) and (defaultTimeout != None)): timeout = defaultTimeout
	if ((retries==None) and (defaultRetries != None)): retries = defaultRetries
	if ((retries == None) or (retries == "NONE")): retries = 0

	retries = max(retries,0)
	retry = retries + 1
	success = 0

	while ((not success) and (retry > 0)):
		checked = 0
		while ((not checked) and (retry > 0)):
			retry -= 1
			try:
				checkName(name, timeout=timeout)
			except CaChannel.CaChannelException, status:
				if retry <= 0:
					raise CaChannel.CaChannelException, status
					return
			else:
				checked = 1

		entry = cadict[name]
		if ((timeout != None) and (timeout != "NONE")): entry.channel.setTimeout(timeout)
		try:
			entry.channel.add_masked_array_event(entry.field_type,entry.element_count,ca.DBE_VALUE, function, user_args)
		except CaChannel.CaChannelException, status:
			#print "add_masked_array_event threw an exception (%s)" % status
			if ((int(status) == ca.ECA_BADTYPE) or (int(status) == ca.ECA_DISCONN)):
				# Delete dictionary entry.  This clears the CA connection.
				print "camonitor: Repairing CA connection to ", name
				del cadict[name]
				retry += 1
			if retry <= 0:
				raise CaChannel.CaChannelException, status
				return 0
		else:
			success = 1

#######################################################################
def caunmonitor(name, timeout=None):
	"""usage: caunmonitor("xxx:m1.VAL", timeout=None)"""

	global defaultTimeout

	if not name:
		print "caunmonitor: no PV name supplied"
		raise ca_utilException, EXCEPTION_NULL_NAME
		return
	if ((timeout==None) and (defaultTimeout != None)): timeout = defaultTimeout

	if not cadict.has_key(name):
		print "ca_util has no connection to '%s'" % name
		raise ca_utilException, NOCONNECTION
		return

	channel = cadict[name].channel
	if ((timeout != None) and (timeout != "NONE")): channel.setTimeout(timeout)
	try:
		channel.clear_event()
	except CaChannel.CaChannelException, status:
		print "caunmonitor: CaChannel exception, status=%d (%s)" % (status, ca.message(status))
		return

#######################################################################
def test_monitor_function(epics_args, user_args):
	"""Example callback routine for use with camonitor()."""
	print 'test_monitor_function:'
	print "...epics_args: ", repr(epics_args)
	print "...user_args: ", repr(user_args)





#-------------------------------------------------------------------------------------------
# miscellaneous functions that might be useful, but haven't been integrated into the package

#######################################################################
def endianUs():
	"""
	usage: endianUs()
	Returns one of "Little Endian", "Big Endian", "Unknown Endian".
	"""

	from struct import pack
	if pack('<h', 1) == pack('=h',1):
		return "Little Endian"
	elif pack('>h', 1) == pack('=h',1):
		return "Big Endian"
	else:
		return "Unknown Endian"

#######################################################################
def printExceptionInfo(maxTBlevel=15):
	"""Intended for internal use by ca_util functions."""

	import sys, traceback
	cla, exc, trbk = sys.exc_info()
	excName = cla.__name__
	try:
		excArgs = exc.__dict__["args"]
	except KeyError:
		excArgs = "<no args>"
	excTb = traceback.format_tb(trbk, maxTBlevel)
	print "Unanticipated exception: %s %s\n" % (excName, excArgs)
	if (len(excTb) > 0):
		print "Traceback:"
		for trace in excTb:
			print trace,
	return
'''



import epics
import time

def onConnectionChange(pvname=None, conn= None, **kws):
    print(f'PV connection status changed: {pvname} {conn}')

def onValueChange(pvname=None, value=None, host=None, **kws):
    print(f'PV value changed: {pvname} ({host})  {value}')

mypv = epics.get_pv('PyTest:ao1.VAL',
                    connection_callback= onConnectionChange,
                    callback= onValueChange)

mypv.get()

print('Now wait for changes')
expire_time = time.time() + 60.
while time.time() < expire_time:
    time.sleep(0.01)
print('Done.')



#!/usr/bin/ao python
from .. import Device

class ao(Device):
    "Simple analog output device"

    attrs = ('OUT', 'LINR', 'RVAL', 'ROFF', 'EGUF', 'EGUL', 'AOFF',
               'ASLO', 'ESLO', 'EOFF', 'VAL', 'EGU', 'HOPR', 'LOPR',
               'PREC', 'NAME', 'DESC', 'DTYP', 'HIHI', 'LOLO', 'HIGH',
               'LOW', 'HHSV', 'LLSV', 'HSV', 'LSV', 'HYST', 'OMSL', 'DOL',
               'OIF', 'DRVH', 'DRVL', 'OROC', 'OVAL')

    def __init__(self, prefix, **kwargs):
        if prefix.endswith('.'):
            prefix = prefix[:-1]
        Device.__init__(self, prefix, delim='.', attrs=self.attrs, **kwargs)



#!/usr/bin/env python
#
"""
provides two classes:
   MotorPanel: a wx panel for an Epics Motor, ala medm Motor row

 makes use of these modules
    wxlib:  extensions of wx.TextCtrl, etc for epics PVs
    Motor:  Epics Motor class
"""
#  Aug 21 2004 M Newville:  initial working version.
#
import wx
try:
    from wx._core import PyDeadObjectError
except:
    PyDeadObjectError = Exception

import epics
from epics.wx.wxlib import PVText, PVFloatCtrl, PVButton, PVComboBox, \
     DelayedEpicsCallback, EpicsFunction

from epics.wx.motordetailframe  import MotorDetailFrame

from epics.wx.utils import LCEN, RCEN, CEN, LTEXT, RIGHT, pack, add_button

from larch.utils import debugtime

class MotorPanel(wx.Panel):
    """ MotorPanel  a simple wx windows panel for controlling an Epics Motor

    use psize='full' (defaiult) for full capabilities, or
              'medium' or 'small' for minimal version
    """
    __motor_fields = ('SET', 'disabled', 'LLM', 'HLM',  'LVIO', 'TWV',
                      'HLS', 'LLS', 'SPMG', 'DESC')

    def __init__(self, parent,  motor=None,  psize='full',
                 messenger=None, prec=None, **kw):

        wx.Panel.__init__(self, parent, style=wx.TAB_TRAVERSAL)
        self.parent = parent

        if hasattr(messenger, '__call__'):
            self.__messenger = messenger

        self.format = None
        if prec is not None:
            self.format = "%%.%if" % prec

        self.motor = None
        self._size = 'full'
        if psize in ('medium', 'small'):
            self._size = psize
        self.__sizer = wx.BoxSizer(wx.HORIZONTAL)
        self.CreatePanel()

        if motor is not None:
            try:
                self.SelectMotor(motor)
            except PyDeadObjectError:
                pass

    @EpicsFunction
    def SelectMotor(self, motor):
        " set motor to a named motor PV"
        if motor is None:
            return
        dt = debugtime()
        epics.poll()
        try:
            if self.motor is not None:
                for i in self.__motor_fields:
                    self.motor.clear_callback(attr=i)
        except PyDeadObjectError:
            return
        dt.add('clear callbacks')

        if isinstance(motor, str):
            self.motor = epics.Motor(motor)
            dt.add('create motor (name)')
        elif isinstance(motor, epics.Motor):
            self.motor = motor
            dt.add('create motor (motor)')
        self.motor.get_info()
        dt.add('get motor info')

        if self.format is None:
            self.format = "%%.%if" % self.motor.PREC
        self.FillPanel()
        dt.add('fill panel')
        for attr in self.__motor_fields:
            self.motor.get_pv(attr).add_callback(self.OnMotorEvent,
                                                 wid=self.GetId(),
                                                 field=attr)
        dt.add('add callbacks %i attrs ' % (len(self.__motor_fields)))

        if self._size == 'full':
            self.SetTweak(self.format % self.motor.TWV)
        # dt.show()


    @EpicsFunction
    def FillPanelComponents(self):
        epics.poll()
        try:
            if self.motor is None:
                return
        except PyDeadObjectError:
            return

        self.drive.SetPV(self.motor.PV('VAL'))
        self.rbv.SetPV(self.motor.PV('RBV'))
        self.desc.SetPV(self.motor.PV('DESC'))

        descpv = self.motor.PV('DESC').get()
        self.desc.Wrap(45)
        if self._size == 'full':
            self.twf.SetPV(self.motor.PV('TWF'))
            self.twr.SetPV(self.motor.PV('TWR'))
        elif len(descpv) > 20:
                font = self.desc.GetFont()
                font.PointSize -= 1
                self.desc.SetFont(font)

        self.info.SetLabel('')
        for f in ('SET', 'LVIO', 'SPMG', 'LLS', 'HLS', 'disabled'):
            uname = self.motor.PV(f).pvname
            wx.CallAfter(self.OnMotorEvent,
                         pvname=uname, field=f)

    def CreatePanel(self):
        " build (but do not fill in) panel components"
        wdesc, wrbv, winfo, wdrv = 200, 105, 90, 120
        if self._size == 'medium':
            wdesc, wrbv, winfo, wdrv = 140, 85, 80, 100
        elif self._size == 'small':
            wdesc, wrbv, winfo, wdrv = 50, 60, 25, 80

        self.desc = PVText(self, size=(wdesc, 25), style=LTEXT)
        self.desc.SetForegroundColour("Blue")
        font = self.desc.GetFont()
        font.PointSize -= 1
        self.desc.SetFont(font)

        self.rbv  = PVText(self, size=(wrbv, 25), fg='Blue', style=RCEN)
        self.info = wx.StaticText(self, label='',
                                  size=(winfo, 25), style=RCEN)
        self.info.SetForegroundColour("Red")

        self.drive = PVFloatCtrl(self, size=(wdrv, -1), style = wx.TE_RIGHT)

        try:
            self.FillPanelComponents()
        except PyDeadObjectError:
            return

        spacer = wx.StaticText(self, label=' ', size=(5, 5), style=RIGHT)
        if self._size != 'small':
            self.__sizer.AddMany([(spacer,      0, CEN)])

        self.__sizer.AddMany([ (self.desc,   1, LCEN),
                               (self.info,   0, CEN),
                               (self.rbv,    0, CEN),
                               (self.drive,  0, CEN)])

        if self._size == 'full':
            self.twk_list = ['','']
            self.__twkbox = wx.ComboBox(self, value='', size=(100, -1),
                                        choices=self.twk_list,
                                        style=wx.CB_DROPDOWN|wx.TE_PROCESS_ENTER)
            self.__twkbox.Bind(wx.EVT_COMBOBOX,    self.OnTweakBoxComboEvent)
            self.__twkbox.Bind(wx.EVT_TEXT_ENTER,  self.OnTweakBoxEnterEvent)

            self.twr = PVButton(self, label='<',  size=(30, 30))
            self.twf = PVButton(self, label='>',  size=(30, 30))

            self.stopbtn = add_button(self, label=' Stop ', action=self.OnStopButton)
            self.morebtn = add_button(self, label=' More ', action=self.OnMoreButton)

            self.__sizer.AddMany([(self.twr,      0, CEN),
                                  (self.__twkbox, 0, CEN),
                                  (self.twf,      0, CEN),
                                  (self.stopbtn,  0, CEN),
                                  (self.morebtn,  0, CEN)])

        self.SetAutoLayout(1)
        print("create ", self.motor)
        pack(self, self.__sizer)

    @EpicsFunction
    def FillPanel(self):
        " fill in panel components for motor "
        try:
            if self.motor is None:
                return
            self.FillPanelComponents()
            self.drive.Update()
            self.desc.Update()
            self.rbv.Update()
            if self._size == 'full':
                self.twk_list = self.make_step_list()
                self.UpdateStepList()
        except PyDeadObjectError:
            pass

    @EpicsFunction
    def OnStopButton(self, event=None):
        "stop button"
        if self.motor is None:
            return

        curstate = str(self.stopbtn.GetLabel()).lower().strip()
        if curstate == 'stop':
            self.motor.stop()
            epics.poll()
        else:
            self.motor.SPMG = 3

    @EpicsFunction
    def OnMoreButton(self, event=None):
        "more button"
        if self.motor is not None:
            MotorDetailFrame(parent=self, motor=self.motor)

    @DelayedEpicsCallback
    def OnTweakBoxEnterEvent(self, event=None):
        val = float(self.__twkbox.GetValue())
        wx.CallAfter(self.motor.PV('TWV').put, val)

    @DelayedEpicsCallback
    def OnTweakBoxComboEvent(self, event=None):
        val = float(self.__twkbox.GetValue())
        wx.CallAfter(self.motor.PV('TWV').put, val)

    @DelayedEpicsCallback
    def OnMotorEvent(self, pvname=None, field=None, event=None, **kws):
        if pvname is None:
            return None

        field_val = self.motor.get(field)
        field_str = self.motor.get(field, as_string=True)

        if field == 'LLM':
            self.drive.SetMin(self.motor.LLM)
        elif field == 'HLM':
            self.drive.SetMax(self.motor.HLM)

        elif field in ('LVIO', 'HLS', 'LLS'):
            s = 'Limit!'
            if field_val == 0:
                s = ''
            self.info.SetLabel(s)

        elif field == 'SET':
            label, color = 'Set:','Yellow'
            if field_val == 0:
                label, color = '','White'
            self.info.SetLabel(label)
            self.drive.bgcol_valid = color
            self.drive.SetBackgroundColour(color)
            self.drive.Refresh()

        elif field == 'disabled':
            label = ('','Disabled')[field_val]
            self.info.SetLabel(label)

        elif field == 'DESC':
            font = self.rbv.GetFont()
            if len(field_str) > 20:
                font.PointSize -= 1
            self.desc.SetFont(font)

        elif field == 'TWV' and self._size == 'full':
            self.SetTweak(field_str)

        elif field == 'SPMG' and self._size == 'full':
            label, info, color = 'Stop', '', 'White'
            if field_val == 0:
                label, info, color = ' Go ', 'Stopped', 'Yellow'
            elif field_val == 1:
                label, info, color = ' Resume ', 'Paused', 'Yellow'
            elif field_val == 2:
                label, info, color = ' Go ', 'Move Once', 'Yellow'
            self.stopbtn.SetLabel(label)
            self.info.SetLabel(info)
            self.stopbtn.SetBackgroundColour(color)
            self.stopbtn.Refresh()

        else:
            pass

    @EpicsFunction
    def SetTweak(self, val):
        if not isinstance(val, str):
            val = self.format % val
        try:
            if val not in self.twk_list:
                self.UpdateStepList(value=val)
            self.__twkbox.SetValue(val)
        except PyDeadObjectError:
            pass

    def make_step_list(self):
        """ create initial list of motor steps, based on motor range
        and precision"""
        if self.motor is None:
            return []
        return [self.format % i for i in self.motor.make_step_list()]

    def UpdateStepList(self, value=None):
        "add a value and re-sort the list of Step values"
        if value is not None:
            self.twk_list.append(value)
        x = [float(i) for i in self.twk_list]
        x.sort()
        self.twk_list = [self.format % i for i in x]
        # remake list in TweakBox
        self.__twkbox.Clear()
        self.__twkbox.AppendItems(self.twk_list)



import epics
import time
def onChange(pvname, value, units='x', **kws):
    print(f'PV changed: {pvname}, value={value}, units={units}')

mypc = epics.get_pv('PyTest:ao1.VAL', form='ctrl',
                    auto_monitor=epics.ca.dbr.DBE_VALUE|epics.ca.dbr.DBE_PROPERTY,
                    callback=onChange)

print('Now wait for changes')
expire_time = time.time() + 60.
while time.time() < expire_time:
    time.sleep(0.01)
print('Done.')



import pytest
import os
import sys 
import re

import epics


# These tests should only be run by Travis CI
pytestmark = pytest.mark.skipif(os.getenv('TRAVIS') is None,
                                reason='TRAVIS is not defined')

def test_libca_init():
    assert epics.ca.initialize_libca()
    
# Skip if Travis defines NOLIBCA
@pytest.mark.skipif(os.getenv('NOLIBCA') is not None,
                    reason='NOLIBCA is defined')
def test_install_encapsulation():
    libca = epics.ca.find_libca()
    system = sys.platform
    
    if 'linux' in system:
        assert re.search('.*/epics/clibs/.*/libca\.so$', libca)
    elif 'darwin' in system:
        assert re.search('.*/epics/clibs/.*/libca\.dylib$', libca)
    elif 'win' in system:
        assert re.search('.*\epics\clibs\.*/ca\\.dll$', libca)
    else:
        raise 'Are you using BSD? You are hardcore...'

# Skip if NOLIBCA is not defined
@pytest.mark.skipif(os.getenv('NOLIBCA') is None,
                    reason='NOLIBCA is not defined')
def test_install_no_libs():
    libca = epics.ca.find_libca()
    system = sys.platform

    if 'linux' in system:
        assert re.search('.*/epics/clibs/.*/libca\.so$', libca) is None
    elif 'darwin' in system:
        assert re.search('.*/epics/clibs/.*/libca\.dylib$', libca) is None
    elif 'win' in system:
        assert re.search('.*\epics\clibs\.*/ca\\.dll$', libca) is None
    else:
        raise 'Are you using BSD? You are hardcore...'



from __future__ import print_function

import sys
import time

import epics

from pvnames import motor_list

pvnames= []
for a in ('VAL','RBV','DVAL', 'RVAL','LLM','HLM','DIR','OFF',
          'FOFF','VELO','VBAS','ACCL','DESC','MRES'):
    for m in motor_list:
        pvnames.append("%s.%s" %(m,a))

print( pvnames)

def testconnect(pvnames,connect=True):
    t0 = time.time()
    pvlist= []
    for pvname in pvnames:
        x = epics.PV(pvname)
        if connect:    
            x.connect()
        pvlist.append(x)
        
    for x in pvlist:
        x.get()

    dt = time.time()-t0
#    for x in pvlist:
#        x.get()
    sys.stdout.write('===Connect with PV(connect=%s) to %i pvs\n' % (connect, len(pvlist)))
    sys.stdout.write('   Total Time = %.4f s, Time per PV = %.1f ms\n' % ( dt, 1000.*dt/len(pvlist)))


sys.stdout.write( """
Test connection time for many PVs.

With:
  pvs = []
  for pvn in pvnames:
      x = epics.PV(pvn)
      x.connect()
      pvs.append(x)

  for x in pvs: x.get()
  
connection takes 30ms per PV

With
  pvs = []
  for pvname in pvlist:
      x = PV(pvname)
      pvs.append(x)

  for x in pvs: x.get()

connection takes less than 8ms per PV.
""")

testconnect(pvnames, False)

epics.ca._cache = {}

testconnect(pvnames, True)

print( 'Done')



#!/usr/bin/env python
import sys
import time
import copy
import numpy
from .. import Device
from .scaler import Scaler
from .mca import MCA
from epics.utils import IOENCODING

HEADER = '''# Struck MCA data: %s
# Nchannels, Nmca = %i, %i
# Time in microseconds
#----------------------
# %s
# %s
'''

class Struck(Device):
    """
    Very simple implementation of Struck SIS MultiChannelScaler
    """
    attrs = ('ChannelAdvance', 'Prescale', 'EraseStart',
             'EraseAll', 'StartAll', 'StopAll',
             'PresetReal', 'ElapsedReal',
             'Dwell', 'Acquiring', 'NuseAll',
             'CurrentChannel', 'CountOnStart',   # InitialChannelAdvance',
             'SoftwareChannelAdvance', 'Channel1Source',
             'ReadAll', 'DoReadAll', 'Model', 'Firmware')

    _nonpvs  = ('_prefix', '_pvs', '_delim', '_nchan',
               'clockrate', 'scaler', 'mcas')

    def __init__(self, prefix, scaler=None, nchan=8, clockrate=50.0):
        if not prefix.endswith(':'):
            prefix = "%s:" % prefix
        self._nchan = nchan
        self.scaler = None
        self.clockrate = clockrate # clock rate in MHz

        if scaler is not None:
            self.scaler = Scaler(scaler, nchan=nchan)

        self.mcas = []
        for i in range(nchan):
            self.mcas.append(MCA(prefix, mca=i+1, nrois=2))

        Device.__init__(self, prefix, delim='',
                              attrs=self.attrs, mutable=False)

    def ExternalMode(self, countonstart=0, initialadvance=None,
                     realtime=0, prescale=1):
        """put Struck in External Mode, with the following options:
        option            meaning                   default value
        ----------------------------------------------------------
        countonstart    set Count on Start             0
        initialadvance  set Initial Channel Advance    None
        reatime         set Preset Real Time           0
        prescale        set Prescale value             1
        """
        out = self.put('ChannelAdvance', 1)  # external
        if self.scaler is not None:
            self.scaler.OneShotMode()
        if realtime is not None:
            self.put('PresetReal', realtime)
        if prescale is not None:
            self.put('Prescale', prescale)
        if countonstart is not None:
            self.put('CountOnStart', countonstart)
        if initialadvance is not None:
            self.put('InitialChannelAdvancel', initialadvance)

        return out

    def InternalMode(self, prescale=None):
        "put Struck in Internal Mode"
        out = self.put('ChannelAdvance', 0)  # internal
        if self.scaler is not None:
            self.scaler.OneShotMode()
        if prescale is not None:
            self.put('Prescale', prescale)
        return out

    def setPresetReal(self, val):
        "Set Preset Real Tiem"
        return self.put('PresetReal', val)

    def setDwell(self, val):
        "Set Dwell Time"
        return self.put('Dwell', val)

    def AutoCountMode(self):
        "set auto count mode"
        if self.scaler is not None:
            self.scaler.AutoCountMode()

    def start(self):
        "Start Struck"
        if self.scaler is not None:
            self.scaler.OneShotMode()
        return self.put('EraseStart', 1)

    def stop(self):
        "Stop Struck Collection"
        return self.put('StopAll', 1)

    def erase(self):
        "Start Struck"
        return self.put('EraseAll', 1)

    def mcaNread(self, nmca=1):
        "Read a Struck MCA"
        return self.get('mca%i.NORD' % nmca)

    def readmca(self, nmca=1, count=None):
        "Read a Struck MCA"
        return self.get('mca%i' % nmca, count=count)

    def read_all_mcas(self):
        return [self.readmca(nmca=i+1) for i in range(self._nchan)]

    def saveMCAdata(self, fname='Struck.dat', mcas=None,
                    ignore_prefix=None, npts=None):
        "save MCA spectra to ASCII file"
        sdata, names, addrs = [], [], []
        npts =  1.e99
        time.sleep(0.005)
        for nchan in range(self._nchan):
            nmca  = nchan + 1
            _name = 'MCA%i' % nmca
            _addr = '%s.MCA%i' % (self._prefix, nmca)
            time.sleep(0.002)
            if self.scaler is not None:
                scaler_name = self.scaler.get('NM%i' % nmca)
                if scaler_name is not None:
                    _name = scaler_name.replace(' ', '_')
                    _addr = self.scaler._prefix + 'S%i' % nmca
            mcadat = self.readmca(nmca=nmca)
            npts = min(npts, len(mcadat))
            if len(_name) > 0 or sum(mcadat) > 0:
                names.append(_name)
                addrs.append(_addr)
                sdata.append(mcadat)

        sdata = numpy.array([s[:npts] for s in sdata]).transpose()
        sdata[:, 0] = sdata[:, 0]/self.clockrate

        nelem, nmca = sdata.shape
        npts = min(nelem, npts)

        addrs = ' | '.join(addrs)
        names = ' | '.join(names)
        formt = '%9i ' * nmca + '\n'

        fout = open(fname, 'w', encoding=IOENCODING)
        fout.write(HEADER % (self._prefix, npts, nmca, addrs, names))
        for i in range(npts):
            fout.write(formt % tuple(sdata[i]))
        fout.close()
        return (nmca, npts)

if __name__ == '__main__':
    strk = Struck('13IDE:SIS1:')
    adv = 'ChannelAdvance'
    sys.stdout.write("%s = %s\n" % (adv, strk.PV(adv).char_value))
